"""
Module for creating a unitary operator for encoding any complex vector
into the wavefunction of a quantum state. For example, the input vector
:math:`[a, b, c, d]` would result in the state

.. math::

    a\\vert 00\\rangle + b\\vert01\\rangle
    + c\\vert10\\rangle + d\\vert11\\rangle

"""
import numpy as np
from six.moves import input


def unitary_operator(state_vector):
    """
    Uses QR factorization to create a unitary operator
    that can encode an arbitrary normalized vector into
    the wavefunction of a quantum state.

    Assumes that the state of the input qubits is to be expressed as

    .. math::

        (1, 0, \\ldots, 0)^T

    :param 1d array state_vector: Normalized vector whose length is at least
                                  two and a power of two.
    :return: Unitary operator that encodes state_vector
    :rtype: 2d array
    """

    if not np.allclose([np.linalg.norm(state_vector)], [1]):
        raise ValueError("Vector must be normalized")

    if 2 ** get_bits_needed(len(state_vector)) != len(state_vector):
        raise ValueError("Vector length must be a power of two and at least two")

    mat = np.identity(len(state_vector), dtype=complex)
    for i in range(len(state_vector)):
        mat[i, 0] = state_vector[i]
    U = np.linalg.qr(mat)[0]

    # make sure U|0> = |v>
    zero_state = np.zeros(len(U))
    zero_state[0] = 1
    if np.allclose(U.dot(zero_state), state_vector):
        return U
    else:
        # adjust phase if needed
        return -1 * U


def fix_norm_and_length(vector):
    """
    Create a normalized and zero padded version of vector.

    :param 1darray vector: a vector with at least one nonzero component.
    :return: a vector that is the normalized version of vector,
             padded at the end with the smallest number of 0s necessary
             to make the length of the vector :math:`2^m`
             for some positive integer :math:`m`.
    :rtype: 1darray
    """
    # normalize
    norm_vector = vector / np.linalg.norm(vector)

    # pad with zeros
    num_bits = get_bits_needed(len(vector))
    state_vector = np.zeros(2 ** num_bits, dtype=complex)
    for i in range(len(vector)):
        state_vector[i] = norm_vector[i]

    return state_vector


def get_bits_needed(n):
    """
    Calculates the smallest positive integer :math:`m` for which
    :math:`2^m\geq n`.

    :param int n: A positive integer
    :return: The positive integer :math:`m`, as specified above
    :rtype: int
    """

    assert n > 0, "Inout should be positive"

    num_bits = int(np.ceil(np.log2(n)))
    return max(1, num_bits)


if __name__ == "__main__":
    num_entries = int(input("How long is the input vector? "))
    num_bits = int(np.ceil(np.log2(num_entries)))
    print("Begin entering vector elements below.")
    vector = []
    for i in range(num_entries):
        vector.append(complex(input("Element {0}: ".format(i))))
    print("You entered the following vector: ", vector)
    state_vector = fix_norm_and_length(vector)
    print("The following vector will be encoded: ", state_vector)
    print("The following matrix was generated: ")
    mat = unitary_operator(state_vector)
    print(mat)
    print("The first column of this matrix is: ")
    print(list(mat[:, 0]))




"""
Finding the minimum energy for an Ising problem by QAOA.
"""
import pyquil.api as api
from grove.pyqaoa.qaoa import QAOA
from pyquil.paulis import PauliSum, PauliTerm
from scipy.optimize import minimize
import numpy as np

CXN = api.SyncConnection()


def energy_value(h, J, sol):
    """
    Obtain energy of an Ising solution for a given Ising problem (h,J).

    :param h: External magnectic term of the Ising problem. List.
    :param J: Interaction term of the Ising problem. Dictionary.
    :param sol: Ising solution. List.
    :return: Energy of the Ising string.
    :rtype: Integer or float.

    """
    ener_ising = 0
    for elm in J.keys():
        if elm[0] == elm[1]:
            raise TypeError("""Interaction term must connect two different variables""")
        else:
            ener_ising += J[elm] * int(sol[elm[0]]) * int(sol[elm[1]])
    for i in range(len(h)):
        ener_ising += h[i] * int(sol[i])
    return ener_ising


def print_fun(x):
    print(x)


def ising_trans(x):
    # Transformation to Ising notation
    if x == 1:
        return -1
    else:
        return 1


def ising(h, J, num_steps=0, verbose=True, rand_seed=None, connection=None, samples=None,
          initial_beta=None, initial_gamma=None, minimizer_kwargs=None,
          vqe_option=None):
    """
    Ising set up method

    :param h: External magnectic term of the Ising problem. List.
    :param J: Interaction term of the Ising problem. Dictionary.
    :param num_steps: (Optional.Default=2 * len(h)) Trotterization order for the
                  QAOA algorithm.
    :param verbose: (Optional.Default=True) Verbosity of the code.
    :param rand_seed: (Optional. Default=None) random seed when beta and
                      gamma angles are not provided.
    :param connection: (Optional) connection to the QVM. Default is None.
    :param samples: (Optional. Default=None) VQE option. Number of samples
                    (circuit preparation and measurement) to use in operator
                    averaging.
    :param initial_beta: (Optional. Default=None) Initial guess for beta
                         parameters.
    :param initial_gamma: (Optional. Default=None) Initial guess for gamma
                          parameters.
    :param minimizer_kwargs: (Optional. Default=None). Minimizer optional
                             arguments.  If None set to
                             {'method': 'Nelder-Mead',
                             'options': {'ftol': 1.0e-2, 'xtol': 1.0e-2,
                                        'disp': False}
    :param vqe_option: (Optional. Default=None). VQE optional
                             arguments.  If None set to
                       vqe_option = {'disp': print_fun, 'return_all': True,
                       'samples': samples}
    :return: Most frequent Ising string, Energy of the Ising string, Circuit used to obtain result.
    :rtype: List, Integer or float, 'pyquil.quil.Program'.

    """
    if num_steps == 0:
        num_steps = 2 * len(h)

    n_nodes = len(h)

    cost_operators = []
    driver_operators = []
    for i, j in J.keys():
        cost_operators.append(PauliSum([PauliTerm("Z", i, J[(i, j)]) * PauliTerm("Z", j)]))

    for i in range(n_nodes):
        cost_operators.append(PauliSum([PauliTerm("Z", i, h[i])]))

    for i in range(n_nodes):
        driver_operators.append(PauliSum([PauliTerm("X", i, -1.0)]))

    if connection is None:
        connection = CXN

    if minimizer_kwargs is None:
        minimizer_kwargs = {'method': 'Nelder-Mead',
                            'options': {'ftol': 1.0e-2, 'xtol': 1.0e-2,
                                        'disp': False}}
    if vqe_option is None:
        vqe_option = {'disp': print_fun, 'return_all': True,
                      'samples': samples}

    if not verbose:
        vqe_option['disp'] = None

    qaoa_inst = QAOA(connection, n_nodes, steps=num_steps, cost_ham=cost_operators,
                     ref_hamiltonian=driver_operators, store_basis=True,
                     rand_seed=rand_seed,
                     init_betas=initial_beta,
                     init_gammas=initial_gamma,
                     minimizer=minimize,
                     minimizer_kwargs=minimizer_kwargs,
                     vqe_options=vqe_option)

    betas, gammas = qaoa_inst.get_angles()
    most_freq_string, sampling_results = qaoa_inst.get_string(
        betas, gammas)
    most_freq_string_ising = [ising_trans(it) for it in most_freq_string]
    energy_ising = energy_value(h, J, most_freq_string_ising)
    param_prog = qaoa_inst.get_parameterized_program()
    circuit = param_prog(np.hstack((betas, gammas)))

    return most_freq_string_ising, energy_ising, circuit



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from collections import Counter
from scipy import optimize
import numpy as np
from grove.pyvqe.vqe import VQE
import pyquil.quil as pq
from pyquil.gates import H
from pyquil.paulis import exponential_map, PauliSum
from functools import reduce


class QAOA(object):
    def __init__(self, qvm, n_qubits, steps=1, init_betas=None,
                 init_gammas=None, cost_ham=[],
                 ref_hamiltonian=[], driver_ref=None,
                 minimizer=None, minimizer_args=[],
                 minimizer_kwargs={}, rand_seed=None,
                 vqe_options={}, store_basis=False):
        """
        QAOA object.

        Contains all information for running the QAOA algorthm to find the
        ground state of the list of cost clauses.

        :param qvm: (Connection) The qvm connection to use for the algorithm.
        :param n_qubits: (int) The number of qubits to use for the algorithm.
        :param steps: (int) The number of mixing and cost function steps to use.
                      Default=1.
        :param init_betas: (list) Initial values for the beta parameters on the
                           mixing terms. Default=None.
        :param init_gammas: (list) Initial values for the gamma parameters on the
                            cost function. Default=None.
        :param cost_ham: list of clauses in the cost function. Must be
                    PauliSum objects
        :param ref_hamiltonian: list of clauses in the cost function. Must be
                    PauliSum objects
        :param driver_ref: (pyQuil.quil.Program()) object to define state prep
                           for the starting state of the QAOA algorithm.
                           Defaults to tensor product of \|+> states.
        :param rand_seed: integer random seed for initial betas and gammas
                          guess.
        :param minimizer: (Optional) Minimization function to pass to the
                          Variational-Quantum-Eigensolver method
        :param minimizer_kwargs: (Optional) (dict) of optional arguments to pass to
                                 the minimizer.  Default={}.
        :param minimizer_args: (Optional) (list) of additional arguments to pass to the
                               minimizer. Default=[].
        :param minimizer_args: (Optional) (list) of additional arguments to pass to the
                               minimizer. Default=[].
        :param vqe_options: (optinal) arguents for VQE run.
        :param store_basis: (optional) boolean flag for storing basis states.
                            Default=False.
        """
        self.qvm = qvm
        self.steps = steps
        self.n_qubits = n_qubits
        self.nstates = 2 ** n_qubits
        if store_basis:
            self.states = [np.binary_repr(i, width=self.n_qubits) for i in range(
                           self.nstates)]
        self.betas = init_betas
        self.gammas = init_gammas
        self.vqe_options = vqe_options

        if driver_ref is not None:
            if not isinstance(driver_ref, pq.Program):
                raise TypeError("""Please provide a pyQuil Program object as a
                                   to generate initial state""")
            else:
                self.ref_state_prep = driver_ref
        else:
            ref_prog = pq.Program()
            for i in range(self.n_qubits):
                ref_prog.inst(H(i))
            self.ref_state_prep = ref_prog

        if not isinstance(cost_ham, (list, tuple)):
            raise TypeError("""cost_hamiltonian must be a list of PauliSum
                               objects""")
        if not all([isinstance(x, PauliSum) for x in cost_ham]):
            raise TypeError("""cost_hamiltonian must be a list of PauliSum
                                   objects""")
        else:
            self.cost_ham = cost_ham

        if not isinstance(ref_hamiltonian, (list, tuple)):
            raise TypeError("""cost_hamiltonian must be a list of PauliSum
                               objects""")
        if not all([isinstance(x, PauliSum) for x in ref_hamiltonian]):
            raise TypeError("""cost_hamiltonian must be a list of PauliSum
                                   objects""")
        else:
            self.ref_ham = ref_hamiltonian

        if minimizer is None:
            self.minimizer = optimize.minimize
        else:
            self.minimizer = minimizer
        # minimizer_kwargs initialized to empty dictionary
        if len(minimizer_kwargs) == 0:
            self.minimizer_kwargs = {'method': 'Nelder-Mead',
                                     'options': {'disp': True,
                                                 'ftol': 1.0e-2,
                                                 'xtol': 1.0e-2}}
        else:
            self.minimizer_kwargs = minimizer_kwargs

        self.minimizer_args = minimizer_args

        if rand_seed is not None:
            np.random.seed(rand_seed)
        if self.betas is None:
            self.betas = np.random.uniform(0, np.pi, self.steps)[::-1]
        if self.gammas is None:
            self.gammas = np.random.uniform(0, 2*np.pi, self.steps)

    def get_parameterized_program(self):
        """
        Return a function that accepts parameters and returns a new Quil
        program

        :returns: a function
        """
        cost_para_programs = []
        driver_para_programs = []

        for idx in range(self.steps):
            cost_list = []
            driver_list = []
            for cost_pauli_sum in self.cost_ham:
                for term in cost_pauli_sum.terms:
                    cost_list.append(exponential_map(term))

            for driver_pauli_sum in self.ref_ham:
                for term in driver_pauli_sum.terms:
                    driver_list.append(exponential_map(term))

            cost_para_programs.append(cost_list)
            driver_para_programs.append(driver_list)

        def psi_ref(params):
            """Construct a Quil program for the vector (beta, gamma).

            :param params: array of 2 . p angles, betas first, then gammas
            :return: a pyquil program object
            """
            if len(params) != 2*self.steps:
                raise ValueError("""params doesn't match the number of parameters set
                                    by `steps`""")
            betas = params[:self.steps]
            gammas = params[self.steps:]

            prog = pq.Program()
            prog += self.ref_state_prep
            for idx in range(self.steps):
                for fprog in cost_para_programs[idx]:
                    prog += fprog(gammas[idx])

                for fprog in driver_para_programs[idx]:
                    prog += fprog(betas[idx])

            return prog

        return psi_ref

    def get_angles(self):
        """
        Finds optimal angles with the quantum variational eigensolver method.

        Stored VQE result

        :returns: ([list], [list]) A tuple of the beta angles and the gamma
                  angles for the optimal solution.
        """
        stacked_params = np.hstack((self.betas, self.gammas))
        vqe = VQE(self.minimizer, minimizer_args=self.minimizer_args,
                  minimizer_kwargs=self.minimizer_kwargs)
        cost_ham = reduce(lambda x, y: x + y, self.cost_ham)
        # maximizing the cost function!
        param_prog = self.get_parameterized_program()
        result = vqe.vqe_run(param_prog, cost_ham, stacked_params, qvm=self.qvm,
                             **self.vqe_options)
        self.result = result
        betas = result.x[:self.steps]
        gammas = result.x[self.steps:]
        return betas, gammas

    def probabilities(self, angles):
        """
        Computes the probability of each state given a particular set of angles.

        :param angles: [list] A concatenated list of angles [betas]+[gammas]
        :return: [list] The probabilities of each outcome given those angles.
        """
        if isinstance(angles, list):
            angles = np.array(angles)

        assert angles.shape[0] == 2 * self.steps, "angles must be 2 * steps"

        param_prog = self.get_parameterized_program()
        prog = param_prog(angles)
        wf, _ = self.qvm.wavefunction(prog)
        wf = wf.amplitudes.reshape((-1, 1))
        probs = np.zeros_like(wf)
        for xx in range(2 ** self.n_qubits):
            probs[xx] = np.conj(wf[xx]) * wf[xx]
        return probs

    def get_string(self, betas, gammas, samples=100):
        """
        Compute the most probable string.

        The method assumes you have passed init_betas and init_gammas with your
        pre-computed angles or you have run the VQE loop to determine the
        angles.  If you have not done this you will be returning the output for
        a random set of angles.

        :param betas: List of beta angles
        :param gammas: List of gamma angles
        :param samples: (int, Optional) number of samples to get back from the
                        QVM.
        :returns: tuple representing the bitstring, Counter object from
                  collections holding all output bitstrings and their frequency.
        """
        if samples <= 0 and not isinstance(samples, int):
            raise ValueError("samples variable must be positive integer")
        param_prog = self.get_parameterized_program()
        stacked_params = np.hstack((betas, gammas))
        sampling_prog = param_prog(stacked_params)
        for i in range(self.n_qubits):
            sampling_prog.measure(i, [i])

        bitstring_samples = self.qvm.run_and_measure(sampling_prog,
                                                     range(self.n_qubits),
                                                     trials=samples)
        bitstring_tuples = list(map(tuple, bitstring_samples))
        freq = Counter(bitstring_tuples)
        most_frequent_bit_string = max(freq, key=lambda x: freq[x])
        return most_frequent_bit_string, freq



import pytest
from pyquil.quil import Program

from grove.amplification.oracles import basis_selector_oracle


def test_basis_selector_oracle():
    """Given that the oracle is a 'black box', we shouldn't test its circuits. We should instead
     test that it gives the right result on a basis set of vectors. Currently we aren't implementing
     this behavior in the Grove tests, so we just write tests for code coverage here."""
    prog = Program()
    qubit0 = prog.alloc()
    qubit1 = prog.alloc()
    _ = basis_selector_oracle([0], "0")
    _ = basis_selector_oracle([0, 1], "01")
    _ = basis_selector_oracle([qubit0, qubit1], "11")


def test_bad_input_basis_selector_oracle():
    with pytest.raises(ValueError):
        _ = basis_selector_oracle([], "100")



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""A module containing useful utility quantum Programs.
"""

from scipy.linalg import sqrtm
import numpy as np

from pyquil.gates import STANDARD_GATES
import pyquil.quil as pq


STANDARD_GATE_NAMES = list(STANDARD_GATES.keys())
SQRT_PREFIX = "SQRT"
CONTROL_PREFIX = "C"
NOT_GATE_LABEL = "NOT"
NOT_GATE = np.array([[0, 1], [1, 0]])
ZERO_PROJECTION = np.array([[1, 0], [0, 0]])
ONE_PROJECTION = np.array([[0, 0], [0, 1]])


class ControlledProgramBuilder(object):
    """
    Returns a controlled u gate with n-1 control_qubits. Useful for constructing oracles.

    Uses a number of gates quadratic in the number of qubits,
    and defines a linear number of new gates. (Roots and adjoints of unitary.)
     See A. Barenco, C. Bennett, R. Cleve (1995) `Elementary Gates for Quantum Computation
     <https://arxiv.org/abs/quant-ph/9503016>`_ for more information.


    :param str gate_name: The name of the gate target.
    :return: The controlled gate.
    """

    def __init__(self):
        self.defined_gates = set()
        self.control_qubits = None
        self.target_qubit = None
        self.operation = None
        self.gate_name = None

    def with_controls(self, control_qubits):
        """Sets the qubits to control on for this controlled operation.

        :param list control_qubits: The indices of the qubits to condition the gate on.
        :return: self, with control_qubits set.
        :rtype: ControlledProgramBuilder
        """
        self.control_qubits = control_qubits
        return self

    def with_target(self, target_qubit):
        """Sets the target qubit for this controlled operation.

        :param int or Qubit target: The Qubit or index of the target of the gate.
        :return: self, with target_qubit set.
        :rtype: ControlledProgramBuilder
        """
        self.target_qubit = target_qubit
        return self

    def with_operation(self, operation):
        """Sets the operation that defines the controlled gate.

        :param numpy.ndarray operation: The unitary gate to be controlled, given as a numpy array.
        :return: self, with operation set.
        :rtype: ControlledProgramBuilder
        """
        self.operation = operation
        return self

    def with_gate_name(self, gate_name):
        """Sets the name for the controlled gate, used in constructing and defining sqrts of the
         gate.

        :param String gate_name:
        :return: self, with gate_name set.
        :rtype: ControlledProgramBuilder
        """
        self.gate_name = gate_name
        return self

    @staticmethod
    def format_gate_name(prefix, gate_name):
        """Formats gate_name by surrounding with brackets and prepending prefix. Used in this case
         to recursively build a program while defining new gates.

        :param str prefix: The prefix to prepend.
        :param str gate_name: The gate name to format and extend.
        :return: The formatted gate name.
        """
        formatted_gate_name = prefix + '-' + gate_name
        return formatted_gate_name

    def build(self):
        """Builds this controlled gate.

        :return: The controlled gate, defined by this object.
        :rtype: Program
        """
        self.defined_gates = set(STANDARD_GATE_NAMES)
        prog = self._recursive_builder(self.operation,
                                       self.gate_name,
                                       self.control_qubits,
                                       self.target_qubit)
        return prog

    def _defgate(self, program, gate_name, gate_matrix):
        """Defines a gate named gate_name with matrix gate_matrix in program. In addtion, updates
         self.defined_gates to track what has been defined.

        :param Program program: Pyquil Program to add the defgate and gate to.
        :param str gate_name: The name of the gate to add to program.
        :param numpy.ndarray gate_matrix: The array corresponding to the gate to define.
        :return: the modified Program.
        :retype: Program
        """
        new_program = pq.Program()
        new_program += program
        if gate_name not in self.defined_gates:
            new_program.defgate(gate_name, gate_matrix)
            self.defined_gates.add(gate_name)
        return new_program

    def _recursive_builder(self, operation, gate_name, control_qubits, target_qubit):
        """Helper function used to define the controlled gate recursively. It uses the algorithm in
         the reference above. Namely it recursively constructs a controlled gate by applying a
         controlled square root of the gate, followed by a toffoli gate with len(control_qubits) - 1
         controls, applying the controlled adjoint of the square root, another toffoli with
         len(control_qubits) - 1 controls, and finally another controlled copy of the gate.

        :param numpy.ndarray operation: The matrix for the unitary to be controlled.
        :param String gate_name: The name for the gate being controlled.
        :param Sequence control_qubits: The qubits that are the controls.
        :param Qubit or Int target_qubit: The qubit that the gate should be applied to.
        :return: The intermediate Program being built.
        :rtype: Program
        """
        control_true = np.kron(ONE_PROJECTION, operation)
        control_false = np.kron(ZERO_PROJECTION, np.eye(2, 2))
        control_root_true = np.kron(ONE_PROJECTION, sqrtm(operation, disp=True))
        controlled_gate = control_true + control_false
        controlled_root_gate = control_root_true + control_false
        sqrt_name = self.format_gate_name(SQRT_PREFIX, gate_name)
        controlled_subprogram = pq.Program()
        control_gate = pq.Program()
        # For the base case, we check to see if there is just one control qubit.
        if len(control_qubits) == 1:
            control_name = self.format_gate_name(CONTROL_PREFIX, gate_name)
            control_gate = self._defgate(control_gate, control_name, controlled_gate)
            control_gate.inst((control_name, control_qubits[0], target_qubit))
            return control_gate

        else:
            control_sqrt_name = self.format_gate_name(CONTROL_PREFIX, sqrt_name)
            control_gate = self._defgate(control_gate, control_sqrt_name, controlled_root_gate)
            control_gate.inst((control_sqrt_name, control_qubits[-1], target_qubit))
            # Here we recurse to build a toffoli gate on n - 1 of the qubits.
            n_minus_one_toffoli = self._recursive_builder(NOT_GATE,
                                                          NOT_GATE_LABEL,
                                                          control_qubits[:-1],
                                                          control_qubits[-1])

            # We recurse to build a controlled sqrt of the target_gate, excluding the last control.
            n_minus_one_controlled_sqrt = self._recursive_builder(sqrtm(operation, disp=True),
                                                                  sqrt_name,
                                                                  control_qubits[:-1],
                                                                  target_qubit)
            controlled_subprogram += control_gate
            controlled_subprogram += n_minus_one_toffoli
            controlled_subprogram += control_gate.dagger()
            controlled_subprogram += n_minus_one_toffoli
            controlled_subprogram += n_minus_one_controlled_sqrt
            return controlled_subprogram



import numpy as np
import pytest
from mock import patch
from pyquil.gates import X, Z, H
from pyquil.quil import Program

from grove.amplification.grover import Grover
from grove.tests.utils.utils_for_testing import prog_equality, synthesize_programs

identity_oracle = Program()
"""Does nothing on all inputs."""


@pytest.fixture()
def x_oracle():
    """Applies an X gate to the ancilla on all inputs. Returns the program, and the query Qubit."""
    p = Program()
    qubit = p.alloc()
    p.inst(X(qubit))
    return p, qubit


def test_trivial_grover():
    """Testing that we construct the correct circuit for Grover's Algorithm with one step, and the
     identity_oracle on one qubit.
     """
    trivial_grover = Program()
    qubit0 = trivial_grover.alloc()
    # First we put the input into uniform superposition.
    trivial_grover.inst(H(qubit0))
    # No oracle is applied, so we just apply the diffusion operator.
    trivial_grover.inst(H(qubit0))
    trivial_grover.inst(Z(qubit0))
    trivial_grover.inst(H(qubit0))
    qubits = [qubit0]
    generated_trivial_grover = Grover().oracle_grover(identity_oracle, qubits, 1)
    generated_trivial_grover.synthesize()
    trivial_grover.synthesize()
    assert prog_equality(generated_trivial_grover, trivial_grover)


def test_x_oracle_one_grover(x_oracle):
    """Testing that Grover's algorithm with an oracle that applies an X gate to the query bit works,
     with one iteration."""
    x_oracle_grover = Program()
    qubit0 = x_oracle_grover.alloc()
    qubits = [qubit0]
    oracle, query_qubit = x_oracle
    with patch("pyquil.quilbase.InstructionGroup.alloc") as mock_alloc:
        mock_alloc.return_value = qubit0
    generated_x_oracle_grover = Grover().oracle_grover(oracle, qubits, 1)
    # First we put the input into uniform superposition.
    x_oracle_grover.inst(H(qubit0))
    # Now an oracle is applied.
    x_oracle_grover.inst(X(query_qubit))
    # We now apply the diffusion operator.
    x_oracle_grover.inst(H(qubit0))
    x_oracle_grover.inst(Z(qubit0))
    x_oracle_grover.inst(H(qubit0))
    synthesize_programs(generated_x_oracle_grover, x_oracle_grover)
    assert prog_equality(generated_x_oracle_grover, x_oracle_grover)


def test_x_oracle_two_grover(x_oracle):
    """Testing that Grover's algorithm with an oracle that applies an X gate to the query bit works,
     with two iterations."""
    x_oracle_grover = Program()
    qubit0 = x_oracle_grover.alloc()
    qubits = [qubit0]
    oracle, query_qubit = x_oracle
    with patch("pyquil.quilbase.InstructionGroup.alloc") as mock_alloc:
        mock_alloc.return_value = qubit0
    generated_x_oracle_grover = Grover().oracle_grover(oracle, qubits, 2)
    # First we put the input into uniform superposition.
    x_oracle_grover.inst(H(qubit0))
    # Two iterations.
    for _ in range(2):
        # Now an oracle is applied.
        x_oracle_grover.inst(X(query_qubit))
        # We apply the diffusion operator.
        x_oracle_grover.inst(H(qubit0))
        x_oracle_grover.inst(Z(qubit0))
        x_oracle_grover.inst(H(qubit0))
    synthesize_programs(generated_x_oracle_grover, x_oracle_grover)
    assert prog_equality(generated_x_oracle_grover, x_oracle_grover)


def test_optimal_grover(x_oracle):
    """Testing that Grover's algorithm with an oracle that applies an X gate to the query bit works,
     and defaults to the optimal number of iterations."""
    grover_precircuit = Program()
    qubit0 = grover_precircuit.alloc()
    qubits = [qubit0]
    oracle, query_qubit = x_oracle
    with patch("pyquil.quilbase.InstructionGroup.alloc") as mock_alloc:
        mock_alloc.return_value = qubit0
        generated_one_iter_grover = Grover().oracle_grover(oracle, qubits)
    # First we put the input into uniform superposition.
    grover_precircuit.inst(H(qubit0))
    # We only do one iteration, which is the result of rounding pi * sqrt(N)/4
    iter = Program()

    # An oracle is applied.
    iter.inst(X(query_qubit))
    # We now apply the diffusion operator.
    iter.inst(H(qubit0))
    iter.inst(Z(qubit0))
    iter.inst(H(qubit0))
    one_iter_grover = grover_precircuit + iter
    synthesize_programs(generated_one_iter_grover, one_iter_grover)
    assert prog_equality(generated_one_iter_grover, one_iter_grover)


def test_find_bistring():
    bitstring_map = {"0": 1, "1": -1}
    builder = Grover()
    with patch("pyquil.api.SyncConnection") as qvm:
        expected_bitstring = [0, 1]
        qvm.run_and_measure.return_value = [expected_bitstring, ]
    returned_bitstring = builder.find_bitstring(qvm, bitstring_map)
    prog = builder.grover_circuit
    # Make sure it only defines the one ORACLE gate.
    assert len(prog.defined_gates) == 1
    # Make sure that it produces the oracle we expect.
    assert (prog.defined_gates[0].matrix == np.array([[1, 0], [0, -1]])).all()
    expected_bitstring = "".join([str(bit) for bit in expected_bitstring])
    returned_bitstring = "".join([str(bit) for bit in returned_bitstring])
    assert expected_bitstring == returned_bitstring



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

"""Module for amplitude amplification, for use in algorithms such as Grover's algorithm.

 See G. Brassard, P. Hoyer, M. Mosca (2000) `Quantum Amplitude Amplification and Estimation
 <https://arxiv.org/abs/quant-ph/0005055 arXiv:quant-ph/0005055>`_ for more information.
"""
import numpy as np
import pyquil.quil as pq
from pyquil.gates import H, X, Z, RZ, STANDARD_GATES

from grove.utils.utility_programs import ControlledProgramBuilder

STANDARD_GATE_NAMES = list(STANDARD_GATES.keys())
X_GATE = np.array([[0, 1], [1, 0]])


def amplification_circuit(algorithm, oracle, qubits, num_iter):
    """
    Returns a program that does ``num_iter`` rounds of amplification, given a measurement-less
    algorithm, an oracle, and a list of qubits to operate on.

    :param Program algorithm: A program representing a measurement-less algorithm run on qubits.
    :param Program oracle: An oracle maps any basis vector ``|psi>`` to either ``+|psi>`` or
        ``-|psi>`` depending on whether ``|psi>`` is in the desirable subspace or the undesirable
        subspace.
    :param Sequence qubits: the qubits to operate on
    :param int num_iter: number of iterations of amplifications to run
    :return: The amplified algorithm.
    :rtype: Program
    """
    if num_iter <= 0:
        raise ValueError("num_iter must be greater than zero")

    prog = pq.Program()

    uniform_superimposer = pq.Program().inst([H(qubit) for qubit in qubits])
    prog += uniform_superimposer

    for _ in range(num_iter):
        prog += oracle + algorithm.dagger() + diffusion_program(qubits) + algorithm
    return prog


def diffusion_program(qubits):
    """
    Constructs the diffusion operator used in Grover's Algorithm, acted on both sides by an
    a Hadamard gate on each qubit. Note that this means that the matrix representation of this
    operator is diag(1, -1, ..., -1).

    See C. Lavor, L.R.U. Manssur, and R. Portugal (2003) `Grover's Algorithm: Quantum Database
    Search`_ for more information.

    .. _`Grover's Algorithm: Quantum Database Search`: https://arxiv.org/abs/quant-ph/0301079

    :param qubits: A list of ints corresponding to the qubits to operate on.
                   The operator operates on bistrings of the form
                   ``|qubits[0], ..., qubits[-1]>``.
    """
    diffusion_program = pq.Program()

    if len(qubits) == 1:
        diffusion_program.inst(Z(qubits[0]))
    else:
        diffusion_program.inst([X(q) for q in qubits])
        diffusion_program.inst(H(qubits[-1]))
        diffusion_program.inst(RZ(-np.pi)(qubits[0]))
        diffusion_program += (ControlledProgramBuilder()
                              .with_controls(qubits[:-1])
                              .with_target(qubits[-1])
                              .with_operation(X_GATE)
                              .with_gate_name("NOT").build())
        diffusion_program.inst(RZ(-np.pi)(qubits[0]))
        diffusion_program.inst(H(qubits[-1]))
        diffusion_program.inst([X(q) for q in qubits])
    return diffusion_program



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import numpy as np
from grove.pyqaoa.qaoa import QAOA
from grove.pyqaoa.utils import compare_progs
import pyquil.api as qvm_module
from pyquil.paulis import PauliTerm, PauliSum
from pyquil.gates import X, Y, Z
from pyquil.quil import Program
from pyquil.wavefunction import Wavefunction
from grove.pyvqe.vqe import VQE
from mock import Mock, patch


def test_probabilities():
    p = 1
    n_qubits = 2
    # known set of angles for barbell
    angles = [1.96348709,  4.71241069]
    wf = np.array([-1.17642098e-05 - 1j*7.67538040e-06,
                   -7.67563580e-06 - 1j*7.07106781e-01,
                   -7.67563580e-06 - 1j*7.07106781e-01,
                   -1.17642098e-05 - 1j*7.67538040e-06])
    fakeQVM = Mock(spec=qvm_module.SyncConnection())
    fakeQVM.wavefunction = Mock(return_value=(Wavefunction(wf), 0))
    inst = QAOA(fakeQVM, n_qubits, steps=p,
                rand_seed=42)

    true_probs = np.zeros_like(wf)
    for xx in range(wf.shape[0]):
        true_probs[xx] = np.conj(wf[xx]) * wf[xx]
    probs = inst.probabilities(angles)
    assert isinstance(probs, np.ndarray)
    prob_true = np.zeros((2**inst.n_qubits, 1))
    prob_true[1] = 0.5
    prob_true[2] = 0.5
    assert np.isclose(probs, prob_true).all()


def test_get_angles():
    p = 2
    n_qubits = 2
    fakeQVM = Mock()
    with patch('grove.pyqaoa.qaoa.VQE', spec=VQE) as mockVQEClass:
        inst = mockVQEClass.return_value
        result = Mock()
        result.x = [1.2, 2.1, 3.4, 4.3]
        inst.vqe_run.return_value = result
        MCinst = QAOA(fakeQVM, n_qubits, steps=p,
                      cost_ham=[PauliSum([PauliTerm("X", 0)])])
        betas, gammas = MCinst.get_angles()
        assert betas == [1.2, 2.1]
        assert gammas == [3.4, 4.3]


def test_ref_program_pass():
    ref_prog = Program().inst([X(0), Y(1), Z(2)])
    fakeQVM = Mock(spec=qvm_module.SyncConnection())
    inst = QAOA(fakeQVM, 2, driver_ref=ref_prog)
    param_prog = inst.get_parameterized_program()
    test_prog = param_prog([0, 0])
    compare_progs(ref_prog, test_prog)



import pytest

from grove.bernstein_vazirani import utils as u


def test_bitwise_dot_product():
    with pytest.raises(ValueError):
        _ = u.bitwise_dot_product('11', '1111')

    assert u.bitwise_dot_product('110', '110') == '0'
    assert u.bitwise_dot_product('110', '101') == '1'
    assert u.bitwise_dot_product('010', '101') == '0'


def test_bit_masking():
    with pytest.raises(ValueError):
        _ = u.bitwise_xor('11', '1111')

    bit_string = '101'
    mask_string = '110'
    assert u.bitwise_xor(bit_string, mask_string) == '011'



import numpy as np
from operator import xor

PADDED_BINARY_BIT_STRING = "{0:0{1:0d}b}"


def is_unitary(matrix):
    """
    A helper function that checks if a matrix is unitary.

    :param 2darray matrix: a matrix to test unitarity of
    :return: true if and only if matrix is unitary
    :rtype: bool
    """
    rows, cols = matrix.shape
    if rows != cols:
        return False
    return np.allclose(np.eye(rows), matrix.dot(matrix.T.conj()))


def most_significant_bit(lst):
    """
    A helper function that finds the position of
    the most significant bit in a 1darray of 1s and 0s,
    i.e. the first position where a 1 appears, reading left to right.

    :param 1darray lst: a 1darray of 0s and 1s with at least one 1
    :return: the first position in lst that a 1 appears
    :rtype: int
    """
    return np.argwhere(np.asarray(lst) == 1)[0][0]


def bitwise_xor(bs0, bs1):
    """
    A helper to calculate the bitwise XOR of two bit string

    :param bs0: String of 0's and 1's representing a number in binary representations
    :param bs1: String of 0's and 1's representing a number in binary representations
    :return: String of 0's and 1's representing the XOR between bs0 and bs1
    :rtype: str
    """
    if len(bs0) != len(bs1):
        raise ValueError("Bit strings are not of equal length")
    n_bits = len(bs0)
    return PADDED_BINARY_BIT_STRING.format(xor(int(bs0, 2), int(bs1, 2)), n_bits)


def binary_back_substitute(W, s):
    """
    Perform back substitution on a binary system of equations, i.e. it performs Gauss elimination
    over the field :math:`GF(2)`. It finds an :math:`\\mathbf{x}` such that
    :math:`\\mathbf{\\mathit{W}}\\mathbf{x}=\\mathbf{s}`, where all arithmetic is taken bitwise
    and modulo 2.

    :param 2darray W: A square :math:`n\\times n` matrix of 0s and 1s,
              in row-echelon (upper-triangle) form
    :param 1darray s: An :math:`n\\times 1` vector of 0s and 1s
    :return: The :math:`n\\times 1` vector of 0s and 1s that solves the above
             system of equations.
    :rtype: 1darray
    """
    # iterate backwards, starting from second to last row for back-substitution
    m = np.copy(s)
    n = len(s)
    for row_num in range(n - 2, -1, -1):
        row = W[row_num]
        for col_num in range(row_num + 1, n):
            if row[col_num] == 1:
                m[row_num] = xor(s[row_num], s[col_num])

    return m[::-1]



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from grove.pyvqe.vqe import VQE, parity_even_p
from pyquil.quil import Program
from pyquil.wavefunction import Wavefunction
from pyquil.gates import RX, H, RZ
from pyquil.paulis import PauliSum, PauliTerm
from mock import Mock, MagicMock, patch
import numpy as np
from scipy.linalg import expm
from scipy.optimize import minimize
import funcsigs
from collections import OrderedDict
import pytest


def test_vqe_run():
    """VQE initialized and then minimizer is called to return result. Checks
    correct sequence of execution"""
    def param_prog(alpha):
        return Program([H(0), RZ(alpha)(0)])

    hamiltonian = np.array([[1, 0], [0, -1]])
    initial_param = 0.0

    minimizer = MagicMock(spec=minimize, func_code=minimize.__code__)
    fake_result = Mock()
    fake_result.fun = 1.0
    fake_result.x = [0.0]
    fake_result.status = 0  # adding so we avoid writing to logger
    minimizer.return_value = fake_result

    # not actually called in VQE run since we are overriding minmizer to just
    # return a value. Still need this so I don't try to call the QVM server.
    fake_qvm = Mock(spec=['wavefunction'])

    with patch("funcsigs.signature") as patch_signature:
        func_sigs_fake = MagicMock(spec=funcsigs.Signature)
        func_sigs_fake.parameters.return_value = \
            OrderedDict({
                'fun': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'x0': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'args': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'method': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'jac': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'hess': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'hessp': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'bounds': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'constraints': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'tol': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'callback': funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                'options': funcsigs.Parameter.POSITIONAL_OR_KEYWORD
            })

        patch_signature.return_value = func_sigs_fake
        inst = VQE(minimizer)

        t_result = inst.vqe_run(param_prog, hamiltonian, initial_param, qvm=fake_qvm)
        assert np.isclose(t_result.fun, 1.0)


def test_expectation():
    """expectation() routine can take a PauliSum operator on a matrix.  Check
    this functionality and the return of the correct scalar value"""
    X = np.array([[0, 1], [1, 0]])

    def RX_gate(phi):
        return expm(-1j*phi*X)

    def rotation_wavefunction(phi):
        state = np.array([[1], [0]])
        return RX_gate(phi).dot(state)

    prog = Program([RX(-2.5)(0)])
    hamiltonian = PauliTerm("Z", 0, 1.0)

    minimizer = MagicMock()
    fake_result = Mock()
    fake_result.fun = 1.0
    minimizer.return_value = fake_result

    fake_qvm = Mock(spec=['wavefunction', 'expectation', 'run'])
    fake_qvm.wavefunction.return_value = (Wavefunction(rotation_wavefunction(-2.5)), [0])
    fake_qvm.expectation.return_value = [0.28366219]
    # for testing expectation
    fake_qvm.run.return_value = [[0], [0]]

    inst = VQE(minimizer)
    energy = inst.expectation(prog, PauliSum([hamiltonian]), None, fake_qvm)
    assert np.isclose(energy, 0.28366219)

    hamiltonian = np.array([[1, 0], [0, -1]])
    energy = inst.expectation(prog, hamiltonian, None, fake_qvm)
    assert np.isclose(energy, 0.28366219)

    prog = Program(H(0))
    hamiltonian = PauliSum([PauliTerm('X', 0)])
    energy = inst.expectation(prog, hamiltonian, 2, fake_qvm)
    assert np.isclose(energy, 1.0)


def test_parity_even_p():
    state_index = 11
    marked_qubits = [0, 1]
    with pytest.raises(AssertionError):
        parity_even_p("", marked_qubits)
    assert parity_even_p(state_index, marked_qubits)


if __name__ == "__main__":
    test_expectation()



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import pyquil.api as api
import pyquil.quil as pq
import numpy as np
from collections import Counter
from pyquil.gates import STANDARD_GATES, RX, RY
from pyquil.paulis import PauliTerm, PauliSum
import funcsigs


class OptResults(dict):
    """
    Object for holding optimization results from VQE.
    """
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class VQE(object):
    """
    The Variational-Quantum-Eigensolver algorithm

    VQE is an object that encapsulates the VQE algorithm (functional
    minimization). The main components of the VQE algorithm are a minimizer
    function for performing the functional minimization, a function that takes a
    vector of parameters and returns a pyQuil program, and a
    Hamiltonian of which to calculate the expectation value.

    Using this object:

        1) initialize with `inst = VQE(minimizer)` where `minimizer` is a
        function that performs a gradient free minization--i.e
        scipy.optimize.minimize(. , ., method='Nelder-Mead')

        2) call `inst.vqe_run(variational_state_evolve, hamiltonian,
        initial_parameters)`. Returns the optimal parameters and minimum
        expecation

    :param minimizer: function that minimizes objective f(obj, param). For
                      example the function scipy.optimize.minimize() needs
                      at least two parameters, the objective and an initial
                      point for the optimization.  The args for minimizer
                      are the cost function (provided by this class),
                      initial parameters (passed to vqe_run() method, and
                      jacobian (defaulted to None).  kwargs can be passed
                      in below.
    :param minimizer_args: (list) arguments for minimizer function. Default=None
    :param minimizer_kwargs: (dict) arguments for keyword args.
                              Default=None

    """

    def __init__(self, minimizer, minimizer_args=[], minimizer_kwargs={}):
        self.minimizer = minimizer
        self.minimizer_args = minimizer_args
        self.minimizer_kwargs = minimizer_kwargs
        self.n_qubits = None

    def vqe_run(self, variational_state_evolve, hamiltonian, initial_params,
                gate_noise=None, measurement_noise=None,
                jacobian=None, qvm=None, disp=None, samples=None, return_all=False):
        """
        functional minimization loop.

        :param variational_state_evolve: function that takes a set of parameters
                                        and returns a pyQuil program.
        :param hamiltonian: (PauliSum) object representing the hamiltonian of
                            which to take the expectation value.
        :param initial_params: (ndarray) vector of initial parameters for the
                               optimization
        :param gate_noise: list of Px, Py, Pz probabilities of gate being
                           applied to every gate after each get application
        :param measurement_noise: list of Px', Py', Pz' probabilities of a X, Y
                                  or Z being applied before a measurement.
        :param jacobian: (optional) method of generating jacobian for parameters
                         (Default=None).
        :param qvm: (optional, QVM) forest connection object.
        :param disp: (optional, bool) display level. If True then each iteration
                     expectation and parameters are printed at each
                     optimization iteration.
        :param samples: (int) Number of samples for calculating the expectation
                        value of the operators.  If `None` then faster method
                        ,dotting the wave function with the operator, is used.
                        Default=None.
        :param return_all: (optional, bool) request to return all intermediate
                           parameters determined during the optimization.
        :return: (vqe.OptResult()) object :func:`OptResult <vqe.OptResult>`.
                 The following fields are initialized in OptResult:
                 -x: set of w.f. ansatz parameters
                 -fun: scalar value of the objective function

                 -iteration_params: a list of all intermediate parameter vectors. Only
                                    returned if 'return_all=True' is set as a vqe_run()
                                    option.

                 -expectation_vals: a list of all intermediate expectation values. Only
                                    returned if 'return_all=True' is set as a
                                    vqe_run() option.
        """
        self._disp_fun = disp if disp is not None else lambda x: None
        iteration_params = []
        expectation_vals = []
        self._current_expectation = None
        if samples is None:
            print("""WARNING: Fast method for expectation will be used. Noise
                     models will be ineffective""")

        if qvm is None:
            qvm = api.SyncConnection(
                    gate_noise=gate_noise,
                    measurement_noise=measurement_noise)
        else:
            self.qvm = qvm

        def objective_function(params):
            """
            closure representing the functional

            :param params: (ndarray) vector of parameters for generating the
                           the function of the functional.
            :return: (float) expectation value
            """
            pyquil_prog = variational_state_evolve(params)
            mean_value = self.expectation(pyquil_prog, hamiltonian, samples, qvm)
            self._current_expectation = mean_value  # store for printing
            return mean_value

        def print_current_iter(iter_vars):
            self._disp_fun("\tParameters: {} ".format(iter_vars))
            if jacobian is not None:
                grad = jacobian(iter_vars)
                self._disp_fun("\tGrad-L1-Norm: {}".format(np.max(np.abs(grad))))
                self._disp_fun("\tGrad-L2-Norm: {} ".format(np.linalg.norm(grad)))

            self._disp_fun("\tE => {}".format(self._current_expectation))
            if return_all:
                iteration_params.append(iter_vars)
                expectation_vals.append(self._current_expectation)

        # using self.minimizer
        arguments = funcsigs.signature(self.minimizer).parameters.keys()

        if disp is not None and 'callback' in arguments:
            self.minimizer_kwargs['callback'] = print_current_iter

        args = [objective_function, initial_params]
        args.extend(self.minimizer_args)
        if 'jac' in arguments:
            self.minimizer_kwargs['jac'] = jacobian

        result = self.minimizer(*args, **self.minimizer_kwargs)

        if hasattr(result, 'status'):
            if result.status != 0:
                self._disp_fun("Classical optimization exited with an error index: %i"
                               % result.status)

        results = OptResults()
        if hasattr(result, 'x'):
            results.x = result.x
            results.fun = result.fun
        else:
            results.x = result

        if return_all:
            results.iteration_params = iteration_params
            results.expectation_vals = expectation_vals
        return results

    @staticmethod
    def expectation(pyquil_prog, pauli_sum, samples, qvm):
        """
        Computes the expectation value of pauli_sum over the distribution
        generated from pyquil_prog.

        :param pyquil_prog: (pyQuil program)
        :param pauli_sum: (PauliSum, ndarray) PauliSum representing the
                          operator of which to calculate the expectation value
                          or a numpy matrix representing the Hamiltonian
                          tensored up to the appropriate size.
        :param samples: (int) number of samples used to calculate the
                        expectation value.  If samples is None then the expectation
                        value is calculated by calculating <psi|O|psi> on the
                        QVM.  Error models will not work if samples is None.

        :param qvm: (qvm connection)

        :returns: (float) representing the expectation value of pauli_sum given
                  given the distribution generated from quil_prog.
        """
        if isinstance(pauli_sum, np.ndarray):
            # debug mode by passing an array
            wf, _ = qvm.wavefunction(pyquil_prog)
            wf = np.reshape(wf.amplitudes, (-1, 1))
            average_exp = np.conj(wf).T.dot(pauli_sum.dot(wf)).real
            return average_exp
        else:
            if not isinstance(pauli_sum, (PauliTerm, PauliSum)):
                raise TypeError("pauli_sum variable must be a PauliTerm or"
                                "PauliSum object")

            if isinstance(pauli_sum, PauliTerm):
                pauli_sum = PauliSum([pauli_sum])

            if samples is None:
                operator_progs = []
                operator_coeffs = []
                for p_term in pauli_sum.terms:
                    op_prog = pq.Program()
                    for qindex, op in p_term:
                        op_prog.inst(STANDARD_GATES[op](qindex))
                    operator_progs.append(op_prog)
                    operator_coeffs.append(p_term.coefficient)

                result_overlaps = qvm.expectation(pyquil_prog,
                                                  operator_programs=operator_progs)
                result_overlaps = list(result_overlaps)
                assert len(result_overlaps) == len(operator_progs), """Somehow we
                didn't get the correct number of results back from the QVM"""
                expectation = sum(list(map(lambda x: x[0]*x[1],
                                           zip(result_overlaps, operator_coeffs))))
                return expectation.real
            else:
                if not isinstance(samples, int):
                    raise TypeError("samples variable must be an integer")
                if samples <= 0:
                    raise ValueError("samples variable must be a postive integer")

                # normal execution via fake sampling
                # stores the sum of contributions to the energy from each operator term
                expectation = 0.0
                for j, term in enumerate(pauli_sum.terms):
                    meas_basis_change = pq.Program()
                    qubits_to_measure = []
                    if term.id() == "":
                        meas_outcome = 1.0
                    else:
                        for index, gate in term:
                            qubits_to_measure.append(index)
                            if gate == 'X':
                                meas_basis_change.inst(RY(-np.pi / 2, index))
                            elif gate == 'Y':
                                meas_basis_change.inst(RX(np.pi / 2, index))

                            meas_outcome = \
                                expectation_from_sampling(pyquil_prog + meas_basis_change,
                                                          qubits_to_measure,
                                                          qvm,
                                                          samples)

                    expectation += term.coefficient * meas_outcome

                return expectation.real


def parity_even_p(state, marked_qubits):
    """
    Calculates the parity of elements at indexes in marked_qubits

    Parity is relative to the binary representation of the integer state.

    :param state: The wavefunction index that corresponds to this state.
    :param marked_qubits: The indexes to be considered in the parity sum.
    :returns: A boolean corresponding to the parity.
    """
    assert isinstance(state, int), "{} is not an integer. Must call " \
                                   "parity_even_p with an integer " \
                                   "state.".format(state)
    mask = 0
    for q in marked_qubits:
        mask |= 1 << q
    return bin(mask & state).count("1") % 2 == 0


def expectation_from_sampling(pyquil_program, marked_qubits, qvm, samples):
    """
    Calculation of Z_{i} at marked_qubits

    Given a wavefunctions, this calculates the expectation value of the Zi
    operator where i ranges over all the qubits given in marked_qubits.

    :param pyquil_program: pyQuil program generating some state
    :param marked_qubits: The qubits within the support of the Z pauli
                          operator whose expectation value is being calculated
    :param qvm: A QVM connection.
    :param samples: Number of bitstrings collected to calculate expectation
                    from sampling.
    :returns: The expectation value as a float.
    """
    # construct program to measure
    for qindex in marked_qubits:
        pyquil_program.measure(qindex, qindex)

    bitstring_samples = qvm.run(pyquil_program, range(max(marked_qubits) + 1), trials=samples)
    bitstring_tuples = list(map(tuple, bitstring_samples))

    freq = Counter(bitstring_tuples)

    # perform weighted average
    expectation = 0
    for bitstring, count in freq.items():
        bitstring_int = int("".join([str(x) for x in bitstring[::-1]]), 2)
        if parity_even_p(bitstring_int, marked_qubits):
            expectation += float(count)/samples
        else:
            expectation -= float(count)/samples
    return expectation



import numpy as np
import pytest

from grove.alpha.fermion_transforms.bktransform import BKTransform
from grove.alpha.fermion_transforms.jwtransform import JWTransform

"""
Some tests inspired by:
https://github.com/ProjectQ-Framework/FermiLib/blob/develop/src/fermilib/transforms/_bravyi_kitaev_test.py
"""


def test_hardcoded_transform():
    n_qubits = 16
    bkt = BKTransform(n_qubits)
    x = bkt.kill(9)
    y = bkt.create(9)

    assert str(x) == '0.5j*Z7*Y9*X11*X15 + (0.5+0j)*Z7*Z8*X9*X11*X15'
    assert str(y) == '-0.5j*Z7*Y9*X11*X15 + (0.5+0j)*Z7*Z8*X9*X11*X15'


def test_term_length():
    # create/kill operators are two-term
    n_qubits = 16
    bkt = BKTransform(n_qubits)
    assert len(bkt.create(3)) == 2
    assert len(bkt.kill(3)) == 2

    n_qubits = 7
    bkt = BKTransform(n_qubits)
    assert len(bkt.create(3)) == 2
    assert len(bkt.kill(3)) == 2


def test_throw_errors():
    # throw error when creation outside qubit range
    n_qubits = 16
    bkt = BKTransform(n_qubits)
    with pytest.raises(IndexError):
        bkt.kill(-1)
    with pytest.raises(IndexError):
        bkt.kill(16)
    with pytest.raises(IndexError):
        bkt.kill(17)


def test_locality_invariant():
    # for n_qubits = 2**d, c_j Majorana is always log2(N) + 1 local
    n_qubits = 16
    bkt = BKTransform(n_qubits)
    invariant = np.log2(n_qubits) + 1
    for index in range(n_qubits):
        op = bkt.kill(index)
        op_terms = op.terms
        for term in op_terms:
            coeff = term.coefficient
            #  Identify the c Majorana terms by real
            #  coefficients and check their length.
            if not isinstance(coeff, complex):
                assert len(term) == invariant


@pytest.mark.skip(reason="pyQuil Pauli needs matrix operator / eigenspectrum "
                         "functionality")
def test_eigenspectrum():
    # Jordan-Wigner and Bravyi-Kitaev operators should give same eigenspectrum

    # single number operator
    n_qubits = 16
    bkt = BKTransform(n_qubits)
    op_BK = bkt.create(3) * bkt.kill(3)
    jwt = JWTransform()
    op_JW = jwt.create(3) * jwt.kill(3)
    assert np.sort(np.linalg.eigvals(op_BK.matrix())) == \
           np.sort(np.linalg.eigvals(op_JW.matrix()))

    # sum of number operators
    op_BK = 0
    op_JW = 0
    for i in [1, 3, 5]:
        op_BK += bkt.create(i) * bkt.kill(i)
        op_JW += jwt.create(i) * jwt.kill(i)
    assert np.sort(np.linalg.eigvals(op_BK.matrix())) == \
           np.sort(np.linalg.eigvals(op_JW.matrix()))

    # scaled number operator
    op_BK = 3 * bkt.create(3) * bkt.kill(3)
    op_JW = 3 * jwt.create(3) * jwt.kill(3)
    assert np.sort(np.linalg.eigvals(op_BK.matrix())) == \
           np.sort(np.linalg.eigvals(op_JW.matrix()))



from operator import xor

PADDED_BINARY_BIT_STRING = "{0:0{1:0d}b}"


def bitwise_dot_product(bs0, bs1):
    """
    A helper to calculate the bitwise dot-product between two string representing bit-vectors

    :param String bs0: String of 0's and 1's representing a number in binary representations
    :param String bs1: String of 0's and 1's representing a number in binary representations
    :return: 0 or 1 as a string corresponding to the dot-product value
    :rtype: String
    """
    if len(bs0) != len(bs1):
        raise ValueError("Bit strings are not of equal length")
    return str(sum([int(bs0[i]) * int(bs1[i]) for i in range(len(bs0))]) % 2)


def bitwise_xor(bs0, bs1):
    """
    A helper to calculate the bitwise XOR of two bit string

    :param String bs0: String of 0's and 1's representing a number in binary representations
    :param String bs1: String of 0's and 1's representing a number in binary representations
    :return: String of 0's and 1's representing the XOR between bs0 and bs1
    :rtype: String
    """
    if len(bs0) != len(bs1):
        raise ValueError("Bit strings are not of equal length")
    n_bits = len(bs0)
    return PADDED_BINARY_BIT_STRING.format(xor(int(bs0, 2), int(bs1, 2)), n_bits)



import numpy as np
from mock import patch

from grove.bernstein_vazirani.bernstein_vazirani import BernsteinVazirani, create_bv_bitmap


def test_bv_bitmap_generator():
    expected_bit_map = {
        '000': '1',
        '001': '1',
        '010': '0',
        '011': '0',
        '100': '0',
        '101': '0',
        '110': '1',
        '111': '1'
    }
    a = '110'
    b = '1'
    actual_bitmap = create_bv_bitmap(a, b)

    assert actual_bitmap == expected_bit_map


def test_bv_class_with_bitmap():
    bit_map = {
        '000': '1',
        '001': '1',
        '010': '0',
        '011': '0',
        '100': '0',
        '101': '0',
        '110': '1',
        '111': '1'
    }

    with patch("pyquil.api.SyncConnection") as qvm:
        # Need to mock multiple returns as an iterable
        qvm.run_and_measure.side_effect = [
            ([0, 1, 1], ),
            ([1], )
        ]

    bv = BernsteinVazirani()
    bv.run(qvm, bit_map).check_solution()
    bv_a, bv_b = bv.solution
    assert bv_a == '110'
    assert bv_b == '1'


def test_bv_class_with_check_results():
    a = '1011'
    b = '0'

    bit_map = create_bv_bitmap(a, b)

    with patch("pyquil.api.SyncConnection") as qvm:
        # Need to mock multiple returns as an iterable
        qvm.run_and_measure.side_effect = [
            ([1, 1, 0, 1], ),
            ([0], )
        ]

    bv = BernsteinVazirani()
    bv.run(qvm, bit_map).check_solution()
    bv_a, bv_b = bv.solution
    assert bv_a == a
    assert bv_b == b


def test_bv_class_with_return_solution():
    a = '1011'
    b = '0'

    bit_map = create_bv_bitmap(a, b)

    with patch("pyquil.api.SyncConnection") as qvm:
        # Need to mock multiple returns as an iterable
        qvm.run_and_measure.side_effect = [
            ([1, 1, 0, 1], ),
            ([0], )
        ]

    bv = BernsteinVazirani()
    bv_a, bv_b = bv.run(qvm, bit_map).get_solution()
    assert bv_a == a
    assert bv_b == b


def test_bv_unitary_generator():
    expected_transition_dct = {
        '000': '100',
        '001': '101',
        '010': '010',
        '011': '011',
        '100': '000',
        '101': '001',
        '110': '110',
        '111': '111'
    }
    expected_unitary = [
        [0., 0., 0., 0., 1., 0., 0., 0.],
        [0., 0., 0., 0., 0., 1., 0., 0.],
        [0., 0., 1., 0., 0., 0., 0., 0.],
        [0., 0., 0., 1., 0., 0., 0., 0.],
        [1., 0., 0., 0., 0., 0., 0., 0.],
        [0., 1., 0., 0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0., 0., 1., 0.],
        [0., 0., 0., 0., 0., 0., 0., 1.]
    ]

    bit_map = {
        '00': '1',
        '01': '1',
        '10': '0',
        '11': '0',
    }
    actual_unitary, actual_dct = BernsteinVazirani()._compute_unitary_oracle_matrix(bit_map)
    assert expected_transition_dct == actual_dct
    np.testing.assert_equal(actual_unitary, expected_unitary)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for the Deutsch-Jozsa Algorithm.
"""
import numpy as np
import pyquil.quil as pq
from pyquil.gates import X, H, CNOT
from pyquil.job_results import wait_for_job


SWAP_MATRIX = np.array([[1, 0, 0, 0],
                        [0, 0, 1, 0],
                        [0, 1, 0, 0],
                        [0, 0, 0, 1]])
r"""The matrix that performs \alpha\ket{ij}\to\alpha\ket{ji}"""

ORACLE_GATE_NAME = "DEUTSCH_JOZSA_ORACLE"


class DeutschJosza(object):
    def __init__(self):
        self.bit_map = None
        self.unitary_matrix = None
        self.n_qubits = None
        self.n_ancillas = None
        self._qubits = None
        self.computational_qubits = None
        self.deutsch_jozsa_circuit = None

    def is_constant(self, cxn, bitstring_map):
        """Computes whether bitstring_map represents a constant function, given that it is constant
         or balanced. Constant means all inputs map to the same value, balanced means half of the
         inputs maps to one value, and half to the other.

        :param JobConnection cxn: The connection object to the Rigetti cloud to run pyQuil programs.
        :param bitstring_map: A dictionary whose keys are bitstrings, and whose values are bits
         represented as strings.
        :type bistring_map: Dict[String, String]
        :return: True if the bitstring_map represented a constant function, false otherwise.
        :rtype: bool
        """
        self._init_attr(bitstring_map)
        job_result = wait_for_job(cxn.run_and_measure(self.deutsch_jozsa_circuit,
                                                      self.computational_qubits))
        # We are only running a single shot, so we are only interested in the first element.
        returned_bitstring = job_result.result['result'][0]
        bitstring = np.array(returned_bitstring, dtype=int)
        constant = all([bit == 0 for bit in bitstring])
        return constant

    def _init_attr(self, bitstring_map):
        """
        Acts instead of __init__ method to instantiate the necessary Deutsch-Jozsa state.

        :param Dict[String, String] bitstring_map: truth-table of the input bitstring map in
        dictionary format, used to construct the oracle in the Deutsch-Jozsa algorithm.
        :return: None
        :rtype: NoneType
        """
        self.bit_map = bitstring_map
        self.n_qubits = len(list(bitstring_map.keys())[0])
        # We use one extra qubit for making the oracle,
        # and one for storing the answer of the oracle.
        self.n_ancillas = 2
        self._qubits = list(range(self.n_qubits + self.n_ancillas))
        self.computational_qubits = self._qubits[:self.n_qubits]
        self.ancillas = self._qubits[self.n_qubits:]
        self.unitary_matrix = self.unitary_function(bitstring_map)
        self.deutsch_jozsa_circuit = self._construct_deutsch_jozsa_circuit()

    def _construct_deutsch_jozsa_circuit(self):
        """
        Builds the Deutsch-Jozsa circuit. Which can determine whether a function f mapping
        :math:`\{0,1\}^n \to \{0,1\}` is constant or balanced, provided that it is one of them.

        :return: A program corresponding to the desired instance of Simon's Algorithm.
        :rtype: Program
        """
        dj_prog = pq.Program()

        # Put the first ancilla qubit (query qubit) into minus state
        dj_prog.inst(X(self.ancillas[0]), H(self.ancillas[0]))

        # Apply Hadamard, Oracle, and Hadamard again
        dj_prog.inst([H(qubit) for qubit in self.computational_qubits])

        # Build the oracle
        oracle_prog = pq.Program()
        oracle_prog.defgate(ORACLE_GATE_NAME, self.unitary_matrix)

        scratch_bit = self.ancillas[1]
        qubits_for_funct = [scratch_bit] + self.computational_qubits
        oracle_prog.inst(tuple([ORACLE_GATE_NAME] + qubits_for_funct))
        dj_prog += oracle_prog

        # Here the oracle does not leave the computational qubits unchanged, so we use a CNOT to
        # to move the result to the query qubit, and then we uncompute with the dagger.
        dj_prog.inst(CNOT(self._qubits[0], self.ancillas[0]))
        dj_prog += oracle_prog.dagger()
        dj_prog.inst([H(qubit) for qubit in self.computational_qubits])
        return dj_prog

    @staticmethod
    def unitary_function(mappings):
        """
        Creates a unitary transformation that maps each state to the values specified
        in mappings.

        Some (but not all) of these transformations involve a scratch qubit, so room for one is
        always provided. That is, if given the mapping of n qubits, the calculated transformation
        will be on n + 1 qubits, where the 0th is the scratch bit and the return value
        of the function is left in the 1st.

        :param mappings: Dictionary of the mappings of f(x) on all length n bitstrings, e.g.

            >>> {'00': '0', '01': '1', '10': '1', '11': '0'}

        :type mappings: Dict[String, Int]
        :return: ndarray representing specified unitary transformation.
        :rtype: np.ndarray
        """
        num_qubits = int(np.log2(len(mappings)))
        bitsum = sum([int(bit) for bit in mappings.values()])

        # Only zeros were entered
        if bitsum == 0:
            return np.kron(SWAP_MATRIX, np.identity(2 ** (num_qubits - 1)))

        # Half of the entries were 0, half 1
        elif bitsum == 2 ** (num_qubits - 1):
            unitary_funct = np.zeros(shape=(2 ** num_qubits, 2 ** num_qubits))
            index_lists = [list(range(2 ** (num_qubits - 1))),
                           list(range(2 ** (num_qubits - 1), 2 ** num_qubits))]
            for j in range(2 ** num_qubits):
                bitstring = np.binary_repr(j, num_qubits)
                value = int(mappings[bitstring])
                mappings.pop(bitstring)
                i = index_lists[value].pop()
                unitary_funct[i, j] = 1
            return np.kron(np.identity(2), unitary_funct)

        # Only ones were entered
        elif bitsum == 2 ** num_qubits:
            x_gate = np.array([[0, 1], [1, 0]])
            return np.kron(SWAP_MATRIX, np.identity(2 ** (num_qubits - 1))).dot(
                np.kron(x_gate, np.identity(2 ** num_qubits)))
        else:
            raise ValueError("f(x) must be constant or balanced")



"""Class for generating a program that can generate an arbitrary quantum state.
References are available at:

- http://140.78.161.123/digital/\
2016_ismvl_logic_synthesis_quantum_state_generation.pdf
- https://arxiv.org/pdf/quant-ph/0407010.pdf

Note that the algorithm used creates a circuit that begins with a target state
and brings it to the all zero state. Thus, many of this module's functions
involve finding gates to be applied in the reversed circuit.
"""
import numpy as np
import pyquil.quil as pq
from pyquil.api import SyncConnection
from pyquil.gates import *
from six.moves import input


def get_uniformly_controlled_rotation_matrix(k):
    """
    Returns the matrix represented by :math:`M_{ij}` in arXiv:quant-ph/0407010.

    This matrix converts the angles of :math:`k`-fold uniformly
    controlled rotations to the angles of the efficient gate decomposition.

    :param int k: number of control qubits
    :return: the matrix :math:`M_{ij}`
    :rtype: 2darray
    """
    M = np.full((2 ** k, 2 ** k), 2 ** -k)
    for i in range(2 ** k):
        g_i = i ^ (i >> 1)  # Gray code for i
        for j in range(2 ** k):
            M[i, j] *= (-1) ** (bin(j & g_i).count("1"))
    return M


def get_cnot_control_positions(k):
    """
    Returns a list of positions for the controls of the CNOTs used when
    decomposing uniformly controlled rotations, as outlined in
    arXiv:quant-ph/0407010.

    Referencing Fig. 2 in the aforementioned paper, this method
    uses the convention that, going up from the target qubit,
    the control qubits are labelled :math:`1, 2, \ldots, k`,
    where :math:`k` is the number of control qubits.
    The returned list provides the qubit that controls
    each successive CNOT, in order from left to right.

    :param int k: the number of control qubits
    :return: the list of positions of the controls
    :rtype: list
    """
    rotation_cnots = [1, 1]
    for i in range(2, k + 1):
        # algorithm described is to replace the last control
        # with a control to the new qubit
        # and then repeat the sequence twice
        rotation_cnots[-1] = i
        rotation_cnots = rotation_cnots + rotation_cnots
    return rotation_cnots


def get_rotation_parameters(phases, magnitudes):
    """
    Simulates one step of rotations.

    Given lists of phases and magnitudes of the same length :math:`N`,
    such that :math:`N=2^n` for some positive integer :math:`n`,
    finds the rotation angles required for one step of phase and magnitude
    unification.

    :param list phases: real valued phases from :math:`-\\pi` to :math:`\\pi`.
    :param list magnitudes: positive, real value magnitudes such that
                       the sum of the square of each magnitude is
                       :math:`2^{-m}` for some nonnegative integer :math:`m`.
    :return: A tuple t of four lists such that

        - t[0] are the z-rotations needed to unify adjacent pairs of phases
        - t[1] are the y-rotations needed to unify adjacent pairs of magnitudes
        - t[2] are the updated phases after these rotations are applied
        - t[3] are the updated magnitudes after these rotations are applied

    :rtype: tuple
    """
    # will hold the angles for controlled rotations
    # in the phase unification and probability unification steps,
    # respectively
    z_thetas = []
    y_thetas = []

    # will hold updated phases and magnitudes after rotations
    new_phases = []
    new_magnitudes = []

    for i in range(0, len(phases), 2):
        # find z rotation angles
        phi = phases[i]
        psi = phases[i + 1]
        z_thetas.append(phi - psi)

        # update phases after applying such rotations
        kappa = (phi + psi) / 2.
        new_phases.append(kappa)

        # find y rotation angles
        a = magnitudes[i]
        b = magnitudes[i + 1]
        if a == 0 and b == 0:
            y_thetas.append(0)
        else:
            y_thetas.append(
                2 * np.arcsin((a - b) / (np.sqrt(2 * (a ** 2 + b ** 2)))))

        # update magnitudes after applying such rotations
        c = np.sqrt((a ** 2 + b ** 2) / 2.)
        new_magnitudes.append(c)

    return z_thetas, y_thetas, new_phases, new_magnitudes


def get_reversed_unification_program(angles, control_indices,
                                     target, controls, mode):
    """
    Gets the Program representing the reversed circuit
    for the decomposition of the uniformly controlled
    rotations in a unification step.

    If :math:`n` is the number of controls, the indices within control indices
    must range from 1 to :math:`n`, inclusive. The length of control_indices
    and the length of angles must both be :math:`2^n`.

    :param list angles: The angles of rotation in the the decomposition,
                        in order from left to right
    :param list control_indices: a list of positions for the controls of the
                                 CNOTs used when decomposing uniformly
                                 controlled rotations; see
                                 get_cnot_control_positions for labelling
                                 conventions.
    :param int target: Index of the target of all rotations
    :param list controls: Index of the controls, in order from bottom to top.
    :param str mode: The unification mode. Is either 'phase', corresponding
                     to controlled RZ rotations, or 'magnitude', corresponding
                     to controlled RY rotations.
    :return: The reversed circuit of this unification step.
    :rtype: Program
    """
    if mode == 'phase':
        gate = RZ
    elif mode == 'magnitude':
        gate = RY
    else:
        raise ValueError("mode must be \'phase\' or \'magnitude\'")

    reversed_gates = []

    for j in range(len(angles)):
        if angles[j] != 0:
            # angle is negated in conjugated/reversed circuit
            reversed_gates.append(gate(-angles[j], target))
        if len(controls) > 0:
            reversed_gates.append(CNOT(controls[control_indices[j] - 1],
                                       target))

    return pq.Program().inst(reversed_gates[::-1])


def create_arbitrary_state(vector, qubits=None):
    """
    This function makes a program that can generate an arbitrary state.

    Applies the methods described in references above.

    Given a complex vector :math:`\\mathbf{a}` with components :math:`a_i`
    (:math:`i` ranging from :math:`0` to :math:`N-1`),
    produce a program that takes in the state :math:`\\vert 0 \\rangle`
    and outputs the state

    .. math::

        \\sum_{i=0}^{N-1}\\frac{a_i}{\\vert \\mathbf{a}\\vert} \\vert i\\rangle

    where :math:`i` is given in its binary expansion.

    :param 1darray vector: the vector to put into qubit form.
    :param list(int) qubits: Which qubits to encode the vector into.
                             Must contain at least the minimum
                             number of qubits :math:`n` needed for all elements
                             of vector to be present as a coefficient in the
                             final state. If more than :math:`n` are provided,
                             only the first :math:`n` will be used.
                             If no list is provided, the default will be
                             qubits :math:`0, 1, \ldots, n-1`.
    :return: a program that takes in :math:`\\vert 0\\rangle^{\otimes n}`
             and produces a state that represents this vector, as
             described above.
    :rtype: Program
    """
    vec_norm = vector / np.linalg.norm(vector)
    n = max(1, int(np.ceil(np.log2(len(vec_norm)))))  # number of qubits needed

    if qubits is None:
        qubits = range(n)

    N = 2 ** n  # number of coefficients
    while len(vec_norm) < N:
        vec_norm = np.append(vec_norm, 0)  # pad with zeros

    magnitudes = list(map(np.abs, vec_norm))
    phases = list(map(np.angle, vec_norm))

    # matrix that converts angles of uniformly controlled rotation
    # to angles of uncontrolled rotations
    # at first, all qubits except for the 0 qubit act as controls
    M = get_uniformly_controlled_rotation_matrix(n - 1)

    # generate the positions of the controls for the CNOTs
    # at first, all qubits except for the 0 qubit act as controls
    rotation_cnots = get_cnot_control_positions(n - 1)

    # because conceptually this algorithm starts with the end state
    # and makes it into the |0> state,
    # the gates will be applied in reverse order
    # from the circuit given in the paper
    reversed_prog = pq.Program()

    for step in range(n):
        # Will hold reversed program corresponding to this particular step.
        reversed_step_prog = pq.Program()

        # get the y and z rotation angles needed to unify pairs
        # of phases and magnitudes, respectively,
        # as well as the new phases and magnitudes that these
        # rotations will update the states to
        z_thetas, y_thetas, phases, magnitudes = \
            get_rotation_parameters(phases, magnitudes)

        # convert these rotation angles
        # to those for uncontrolled rotations + CNOTs
        converted_z_thetas = np.dot(M, z_thetas)
        converted_y_thetas = np.dot(M, y_thetas)

        # phase unification
        phase_prog = get_reversed_unification_program(converted_z_thetas,
                                                      rotation_cnots,
                                                      qubits[0],
                                                      qubits[step + 1:],
                                                      'phase')

        # probability unification
        prob_prog = get_reversed_unification_program(converted_y_thetas,
                                                     rotation_cnots,
                                                     qubits[0],
                                                     qubits[step + 1:],
                                                     'magnitude')

        # swaps are applied first, then reversed probability unification,
        # then reversed phase unification

        if step < n - 1:
            # swaps are applied after all rotation steps except the last
            reversed_step_prog += SWAP(qubits[0], qubits[step + 1])

            # just retain upper left square
            # for the next iteration (one less control)
            M = M[0:int(len(M) / 2), 0:int(len(M) / 2)] * 2

            # update next set of controls (one less control)
            rotation_cnots = rotation_cnots[:int(len(rotation_cnots) / 2)]
            rotation_cnots[-1] -= 1

        reversed_step_prog += prob_prog
        reversed_step_prog += phase_prog

        # This step must be applied to the front
        # of the currently built up reversed_prog because
        # of the reversal of the steps in the final reversed program.
        reversed_prog = reversed_step_prog + reversed_prog

    # Add Hadamard gates to remove superposition
    reversed_prog = pq.Program().inst(list(map(H, qubits))) + reversed_prog

    # Correct the overall phase
    reversed_prog = pq.Program().inst(RZ(-2 * phases[0], qubits[0])) \
                      .inst(PHASE(2 * phases[0], qubits[0])) + reversed_prog

    # Apply all gates in reverse
    return reversed_prog


if __name__ == "__main__":
    print("Example list: -3.2+1j, -7, -0.293j, 1, 0, 0")
    v = input("Input a comma separated list of complex numbers:\n")
    if isinstance(v, int):
        v = [v]
    else:
        v = list(v)
    p = create_arbitrary_state(v)
    qvm = SyncConnection()
    wf, _ = qvm.wavefunction(p)
    print("Normalized Vector: ", list(v / np.linalg.norm(v)))
    print("Generated Wavefunction: ", wf)
    if input("Show Program? (y/n): ") == 'y':
        print("----------Quil Code Used----------")
        print(p.out())



"""Test class for helper methods found simon"""

from os.path import abspath, dirname

import numpy as np
from mock import patch
from pyquil.quil import Program

from grove.simon.simon import Simon, create_1to1_bitmap, create_valid_2to1_bitmap

package_path = abspath(dirname(dirname(__file__)))

EXPECTED_SIMON_ORACLE = np.load(package_path + '/simon/data/simon_test_oracle.npy')


def _create_expected_program():
    expected_prog = Program()
    expected_prog.defgate("SIMON_ORACLE", EXPECTED_SIMON_ORACLE)
    expected_prog.inst("H 0")
    expected_prog.inst("H 1")
    expected_prog.inst("H 2")

    expected_prog.inst("SIMON_ORACLE 5 4 3 2 1 0")

    expected_prog.inst("H 0")
    expected_prog.inst("H 1")
    expected_prog.inst("H 2")
    return expected_prog


def test_simon_class():
    """Test is based on worked example of Watrous lecture
    https://cs.uwaterloo.ca/~watrous/CPSC519/LectureNotes/06.pdf"""
    simon_algo = Simon()

    with patch("pyquil.api.SyncConnection") as qvm:
        # Need to mock multiple returns as an iterable
        qvm.run_and_measure.side_effect = [
            (np.asarray([1, 1, 1], dtype=int), ),
            (np.asarray([1, 1, 1], dtype=int), ),
            (np.asarray([1, 0, 0], dtype=int), ),
            (np.asarray([1, 1, 1], dtype=int), ),
            (np.asarray([0, 0, 0], dtype=int), ),
            (np.asarray([0, 1, 1], dtype=int), ),
        ]

    bit_string_mapping = {
        '000': '101',
        '001': '010',
        '010': '000',
        '011': '110',

        '100': '000',
        '101': '110',
        '110': '101',
        '111': '010'
    }

    mask = simon_algo.find_mask(qvm, bit_string_mapping)

    assert simon_algo.n_qubits == 3
    assert simon_algo.n_ancillas == 3
    assert simon_algo._qubits == [0, 1, 2, 3, 4, 5]
    assert simon_algo.computational_qubits == [0, 1, 2]
    assert simon_algo.ancillas == [3, 4, 5]

    assert mask == [1, 1, 0]
    assert simon_algo.simon_circuit.__str__() == _create_expected_program().__str__()


def test_unitary_function_return():
    simon_algo = Simon()
    bit_string_mapping = {
        '000': '101',
        '001': '010',
        '010': '000',
        '011': '110',

        '100': '000',
        '101': '110',
        '110': '101',
        '111': '010'
    }

    actual_return = simon_algo._compute_unitary_oracle_matrix(bit_string_mapping)
    np.testing.assert_equal(actual_return[0], EXPECTED_SIMON_ORACLE)


def test_unitary_oracle_func_computer():
    bit_string_mapping = {
            '0': '1',
            '1': '0',
        }
    np.testing.assert_equal(Simon()._compute_unitary_oracle_matrix(bit_string_mapping)[0],
                            [[0., 0., 1., 0.],
                             [0., 1., 0., 0.],
                             [1., 0., 0., 0.],
                             [0., 0., 0., 1.]]
                            )


def test_unitary_oracle_func_computer_2():
    bit_string_mapping = {
            '00': '10',
            '01': '11',
            '10': '00',
            '11': '01'
        }
    np.testing.assert_equal(Simon()._compute_unitary_oracle_matrix(bit_string_mapping)[0],
                            [[0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.],
                             [0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.],
                             [0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.]]
                            )


def test_no_substitution():
    simon_algo = Simon()
    simon_algo._dict_of_linearly_indep_bit_vectors = {
        0: [1, 0, 1, 0, 0],
        1: [0, 1, 0, 0, 0],
        3: [0, 0, 0, 1, 0]
    }
    z = np.array([1, 1, 1, 0, 0])  # linear combination of first two rows hence won't add

    simon_algo._add_to_dict_of_indep_bit_vectors(z)
    W_actual = simon_algo._dict_of_linearly_indep_bit_vectors

    W_expected = {
        0: [1, 0, 1, 0, 0],
        1: [0, 1, 0, 0, 0],
        3: [0, 0, 0, 1, 0]
    }

    np.testing.assert_equal(W_actual, W_expected)


def test_simple_conflict():
    simon_algo = Simon()
    simon_algo._dict_of_linearly_indep_bit_vectors = {
        0: [1, 0, 1, 0, 0],
        1: [0, 1, 0, 0, 0],
        3: [0, 0, 0, 1, 0]
    }
    z = np.array([1, 0, 0, 0, 1])  # conflict with first row.

    simon_algo._add_to_dict_of_indep_bit_vectors(z)
    W_actual = simon_algo._dict_of_linearly_indep_bit_vectors

    W_expected = {
        0: [1, 0, 1, 0, 0],
        1: [0, 1, 0, 0, 0],
        2: [0, 0, 1, 0, 1],
        3: [0, 0, 0, 1, 0]
    }

    np.testing.assert_equal(W_actual, W_expected)


def test_insert_directly():
    simon_algo = Simon()
    simon_algo._dict_of_linearly_indep_bit_vectors = {
        0: [1, 1, 0, 0, 0],
        1: [0, 1, 0, 1, 0]
    }
    z = np.array([0, 0, 1, 0, 1])

    simon_algo._add_to_dict_of_indep_bit_vectors(z)
    W_actual = simon_algo._dict_of_linearly_indep_bit_vectors
    W_expected = {
        0: [1, 1, 0, 0, 0],
        1: [0, 1, 0, 1, 0],
        2: [0, 0, 1, 0, 1]
    }

    np.testing.assert_equal(W_actual, W_expected)


def test_insert_after_xor():
    simon_algo = Simon()
    simon_algo._dict_of_linearly_indep_bit_vectors = {
        0: [1, 0, 0, 0, 0, 0],
        1: [0, 1, 1, 0, 0, 0]
    }

    z = np.array([0, 0, 1, 0, 1, 1])

    simon_algo._add_to_dict_of_indep_bit_vectors(z)
    W_actual = simon_algo._dict_of_linearly_indep_bit_vectors
    W_expected = {
        0: [1, 0, 0, 0, 0, 0],
        1: [0, 1, 1, 0, 0, 0],
        2: [0, 0, 1, 0, 1, 1]
    }

    np.testing.assert_equal(W_actual, W_expected)


def test_add_row_at_top():
    simon_algo = Simon()
    simon_algo.n_qubits = 4
    simon_algo._dict_of_linearly_indep_bit_vectors = {
        1: [0, 1, 0, 1],
        2: [0, 0, 1, 0],
        3: [0, 0, 0, 1]
    }
    insert_row_num = simon_algo._add_missing_msb_vector()

    W_actual = simon_algo._dict_of_linearly_indep_bit_vectors
    W_expected = {
        0: [1, 0, 0, 0],
        1: [0, 1, 0, 1],
        2: [0, 0, 1, 0],
        3: [0, 0, 0, 1]
    }

    assert insert_row_num == 0

    np.testing.assert_equal(W_actual, W_expected)


def test_add_row_at_bottom():
    simon_algo = Simon()
    simon_algo.n_qubits = 4
    simon_algo._dict_of_linearly_indep_bit_vectors = {
        0: [1, 0, 0, 0],
        1: [0, 1, 0, 1],
        2: [0, 0, 1, 0]
    }
    insert_row_num = simon_algo._add_missing_msb_vector()

    W_actual = simon_algo._dict_of_linearly_indep_bit_vectors
    W_expected = {
        0: [1, 0, 0, 0],
        1: [0, 1, 0, 1],
        2: [0, 0, 1, 0],
        3: [0, 0, 0, 1]
    }
    assert insert_row_num == 3

    np.testing.assert_equal(W_actual, W_expected)


def test_add_row_in_middle():
    simon_algo = Simon()
    simon_algo.n_qubits = 5
    simon_algo._dict_of_linearly_indep_bit_vectors = {
        0: [1, 1, 0, 0, 0],
        2: [0, 0, 1, 0, 1],
        3: [0, 0, 0, 1, 0],
        4: [0, 0, 0, 0, 1]
    }
    insert_row_num = simon_algo._add_missing_msb_vector()

    W_actual = simon_algo._dict_of_linearly_indep_bit_vectors
    W_expected = {
        0: [1, 1, 0, 0, 0],
        1: [0, 1, 0, 0, 0],
        2: [0, 0, 1, 0, 1],
        3: [0, 0, 0, 1, 0],
        4: [0, 0, 0, 0, 1]
    }

    assert insert_row_num == 1

    np.testing.assert_equal(W_actual, W_expected)


def test_bit_map_generation():
    mask = '101'
    expected_map = {
        '000': '101',
        '001': '100',
        '010': '111',
        '011': '110',
        '100': '001',
        '101': '000',
        '110': '011',
        '111': '010'
    }
    actual_map = create_1to1_bitmap(mask)
    assert actual_map == expected_map


def test_2to1_bit_map_generation():
    mask = '101'
    expected_map = {
        '000': '001',
        '101': '001',
        '001': '101',
        '100': '101',
        '010': '000',
        '111': '000',
        '011': '111',
        '110': '111'
    }
    # need to patch numpy as random seed behaves differently on
    # py27 vs. py36
    with patch("numpy.random.choice") as rd_fake:
        rd_fake.return_value = ['001', '101', '000', '111']

        actual_map = create_valid_2to1_bitmap(mask)
        assert actual_map == expected_map


def test_check_mask_correct():
    sa = Simon()

    sa.mask = [1, 1, 0]
    sa.bit_map = {
        '000': '101',
        '001': '010',
        '010': '000',
        '011': '110',

        '100': '000',
        '101': '110',
        '110': '101',
        '111': '010'
    }

    assert sa._check_mask_correct()



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

"""
The Jordan-Wigner Transform
"""
from pyquil.paulis import PauliTerm


class JWTransform(object):
    """
    Jordan-Wigner object the appropriate Pauli operators
    """
    def create(self, index):
        """
        Fermion creation operator at orbital 'n'

        :param Int index: creation index
        """
        return self._operator_generator(index, -1.0)

    def kill(self, index):
        """
        Fermion annihilation operator at orbital 'n'

        :param Int index: annihilation index
        """
        return self._operator_generator(index, 1.0)

    def product_ops(self, indices, conjugate):
        """
        Convert a list of site indices and coefficients to a Pauli Operators
        list with the Jordan-Wigner (JW) transformation

        :param List indices: list of ints specifying the site the fermionic operator acts on,
                             e.g. [0,2,4,6]
        :param List conjugate: List of -1, 1 specifying which of the indices are
                               creation operators (-1) and which are annihilation
                               operators (1).  e.g. [-1,-1,1,1]
        """
        pterm = PauliTerm('I', 0, 1.0)
        for conj, index in zip(conjugate, indices):
            pterm = pterm * self._operator_generator(index, conj)

        pterm = pterm.simplify()
        return pterm

    @staticmethod
    def _operator_generator(index, conj):
        """
        Internal method to generate the appropriate operator
        """
        pterm = PauliTerm('I', 0, 1.0)
        Zstring = PauliTerm('I', 0, 1.0)
        for j in range(index):
            Zstring = Zstring*PauliTerm('Z', j, 1.0)

        pterm1 = Zstring*PauliTerm('X', index, 0.5)
        scalar = 0.5 * conj * 1.0j
        pterm2 = Zstring*PauliTerm('Y', index, scalar)
        pterm = pterm * (pterm1 + pterm2)

        pterm = pterm.simplify()
        return pterm



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

"""
Finding a maximum cut by QAOA.
"""
import numpy as np
import pyquil.api as api
from pyquil.paulis import PauliTerm, PauliSum
import networkx as nx
from scipy.optimize import minimize
from grove.pyqaoa.qaoa import QAOA
CXN = api.SyncConnection()


def print_fun(x):
    print(x)


def maxcut_qaoa(graph, steps=1, rand_seed=None, connection=None, samples=None,
                initial_beta=None, initial_gamma=None, minimizer_kwargs=None,
                vqe_option=None):
    """
    Max cut set up method

    :param graph: Graph definition. Either networkx or list of tuples
    :param steps: (Optional. Default=1) Trotterization order for the QAOA algorithm.
    :param rand_seed: (Optional. Default=None) random seed when beta and gamma angles
        are not provided.
    :param connection: (Optional) connection to the QVM. Default is None.
    :param samples: (Optional. Default=None) VQE option. Number of samples
        (circuit preparation and measurement) to use in operator averaging.
    :param initial_beta: (Optional. Default=None) Initial guess for beta parameters.
    :param initial_gamma: (Optional. Default=None) Initial guess for gamma parameters.
    :param minimizer_kwargs: (Optional. Default=None). Minimizer optional arguments.  If None set to
        ``{'method': 'Nelder-Mead', 'options': {'ftol': 1.0e-2, 'xtol': 1.0e-2, 'disp': False}``
    :param vqe_option: (Optional. Default=None). VQE optional arguments.  If None set to
        ``vqe_option = {'disp': print_fun, 'return_all': True, 'samples': samples}``

    """
    if not isinstance(graph, nx.Graph) and isinstance(graph, list):
        maxcut_graph = nx.Graph()
        for edge in graph:
            maxcut_graph.add_edge(*edge)
        graph = maxcut_graph.copy()

    cost_operators = []
    driver_operators = []
    for i, j in graph.edges():
        cost_operators.append(PauliTerm("Z", i, 0.5)*PauliTerm("Z", j) + PauliTerm("I", 0, -0.5))
    for i in graph.nodes():
        driver_operators.append(PauliSum([PauliTerm("X", i, -1.0)]))

    if connection is None:
        connection = CXN

    if minimizer_kwargs is None:
        minimizer_kwargs = {'method': 'Nelder-Mead',
                            'options': {'ftol': 1.0e-2, 'xtol': 1.0e-2,
                                        'disp': False}}
    if vqe_option is None:
        vqe_option = {'disp': print_fun, 'return_all': True,
                      'samples': samples}

    qaoa_inst = QAOA(connection, len(graph.nodes()), steps=steps, cost_ham=cost_operators,
                     ref_hamiltonian=driver_operators, store_basis=True,
                     rand_seed=rand_seed,
                     init_betas=initial_beta,
                     init_gammas=initial_gamma,
                     minimizer=minimize,
                     minimizer_kwargs=minimizer_kwargs,
                     vqe_options=vqe_option)

    return qaoa_inst


if __name__ == "__main__":
    # Sample Run:
    # Cutting 0 - 1 - 2 graph!
    inst = maxcut_qaoa([(0, 1), (1, 2)],
                       steps=2, rand_seed=42, samples=None)
    betas, gammas = inst.get_angles()
    probs = inst.probabilities(np.hstack((betas, gammas)))
    for state, prob in zip(inst.states, probs):
        print(state, prob)

    print("Most frequent bitstring from sampling")
    most_freq_string, sampling_results = inst.get_string(
            betas, gammas)
    print(most_freq_string)



###############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
###############################################################################
"""
The Bravyi-Kitaev transform, generalized using a Fenwick tree for arbitrary
numbers of qubits.

References:
(Regular Bravyi-Kitaev):
https://arxiv.org/abs/1208.5986

(Fenwick generalization for arbitrary n_qubits):
https://journals.aps.org/pra/abstract/10.1103/PhysRevA.95.032332
"""
from pyquil.paulis import PauliTerm
from .fenwick_tree import FenwickTree


class BKTransform(object):
    """
    Transform object to create arbitrary creation/annihilation fermionic 
    operators, over a pre-specified number of qubits.

    Uses a Fenwick tree / binary index tree to generalize the Bravyi-Kitaev
    transform for arbitrary number of qubits/fermionic lattice sites.

    :param int n_qubits: number of fermionic lattice sites / qubits to
                         initialize
    """
    def __init__(self, n_qubits):
        assert n_qubits >= 0

        # Build the Fenwick tree
        self.tree = FenwickTree(n_qubits)
        self.n_qubits = n_qubits

    def create(self, index):
        """
        Fermion creation operator at orbital 'index'

        :param int index: orbital index to create fermion at

        :return: qubit operators corresponding to fermion creation
        :rtype: PauliSum
        """
        return self._operator_generator(index, -1.0)

    def kill(self, index):
        """
        Fermion annihilation operator at orbital 'index'

        :param int index: orbital index to annihilate fermion at

        :return: qubit operators corresponding to fermion annihilation
        :rtype: PauliSum
        """
        return self._operator_generator(index, +1.0)

    def product_ops(self, indices, conjugate):
        """
        Convert a list of site indices and coefficients to a Pauli Operators
        list with the generalized Bravyi-Kitaev transformation.

        :param list indices: list of ints specifying the site the fermionic
                             operator acts on, e.g. [0,2,4,6]
        :param list conjugate: List of -1, 1 specifying which of the indices
                               are creation operators (-1) and which are
                               annihilation operators (1).  e.g. [-1,-1,1,1]
        """
        pterm = PauliTerm('I', 0, 1.0)
        for conj, index in zip(conjugate, indices):
            pterm = pterm * self._operator_generator(index, conj)

        pterm = pterm.simplify()
        return pterm

    def _operator_generator(self, index, conj):
        """
        Internal method to generate the appropriate ladder operator at fermion
        orbital at 'index'
        If conj == -1 --> creation
           conj == +1 --> annihilation

        :param int index: fermion orbital to generate ladder operator at
        :param int conj: -1 for creation, +1 for annihilation
        """
        if conj != -1 and conj != +1:
            raise ValueError("Improper conjugate coefficient")
        if index >= self.n_qubits or index < 0:
            raise IndexError("Operator index outside number of qubits for "
                             "current Bravyi-Kitaev transform.")

        # parity set P(j). apply Z to, for parity sign.
        parity_set = [node.index for node in self.tree.get_parity_set(index)]

        # update set U(j). apply X to, for updating purposes.
        ancestors = [node.index for node in self.tree.get_update_set(index)]

        # remainder set C(j) = P(j) \ F(j)
        ancestor_children = [node.index for node in self.tree.get_remainder_set(index)]

        # Under Majorana basis, creation/annihilation operators given by
        # a^{\pm} = (c \mp id) / 2

        # c_j = a_j + a_j^{\dagger} = X_{U(j)} X_j Z_{P(j)}
        c_maj = PauliTerm('X', index)
        for node_idx in parity_set:
            c_maj *= PauliTerm('Z', node_idx)
        for node_idx in ancestors:
            c_maj *= PauliTerm('X', node_idx)

        # d_j = i(a_j^{\dagger} - a_j) = X_{U(j)} Y_j Z_{C(j)}
        d_maj = PauliTerm('Y', index)
        for node_idx in ancestors:
            d_maj *= PauliTerm('X', node_idx)
        for node_idx in ancestor_children:
            d_maj *= PauliTerm('Z', node_idx)

        result = 0.5 * (c_maj + 1j * conj * d_maj)
        return result.simplify()



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

"""
Module for the Bernstein-Vazirani Algorithm.
For more information, see [Loceff2015]_

.. [Loceff2015] Loceff, M. (2015), `"A Course in Quantum Computing for the Community College"`_,
 Volume 1, Chapter 18, p 484-541.

.. _`"A Course in Quantum Computing for the Community College"`: http://lapastillaroja.net/
 wp-content/uploads/2016/09/Intro_to_QC_Vol_1_Loceff.pdf
"""

from collections import defaultdict

import numpy as np
import pyquil.quil as pq
from pyquil.gates import H, X

from grove.bernstein_vazirani import utils


def create_bv_bitmap(dot_product_vector, dot_product_bias):
    """
    This function creates a map from bitstring to function value for a boolean formula :math:`f`
    with a dot product vector :math:`a` and a dot product bias :math:`b`

        .. math::

           f:\\{0,1\\}^n\\rightarrow \\{0,1\\}

           \\mathbf{x}\\rightarrow \\mathbf{a}\\cdot\\mathbf{x}+b\\pmod{2}

           (\\mathbf{a}\\in\\{0,1\\}^n, b\\in\\{0,1\\})

    :param String dot_product_vector: a string of 0's and 1's that represents the dot-product
        partner in :math:`f`
    :param String dot_product_bias: 0 or 1 as a string representing the bias term in :math:`f`
    :return: A dictionary containing all possible bitstring of length equal to :math:`a` and the
        function value :math:`f`
    :rtype: Dict[String, String]
    """
    n_bits = len(dot_product_vector)
    bit_map = {}
    for bit_val in range(2 ** n_bits):
        bit_map[np.binary_repr(bit_val, width=n_bits)] = str(
            (int(utils.bitwise_dot_product(np.binary_repr(bit_val, width=n_bits),
                                           dot_product_vector))
             + int(dot_product_bias, 2)) % 2
        )

    return bit_map


class BernsteinVazirani(object):
    """
    This class contains an implementation of the Bernstein-Vazirani algorithm using pyQuil. For more
    references see the documentation_

    .. _documentation: http://grove-docs.readthedocs.io/en/latest/bernstein_vazirani.html
    """

    def __init__(self):
        self.n_qubits = None
        self.n_ancillas = 1
        self.computational_qubits = None
        self.ancilla = None
        self.bv_circuit = None
        self.solution = None
        self.input_bitmap = None

    @staticmethod
    def _compute_unitary_oracle_matrix(bitstring_map):
        """
        Computes the unitary matrix that encodes the oracle function  used in the Bernstein-Vazirani
        algorithm. It generates a dense matrix for a function :math:`f`

        .. math::

           f:\\{0,1\\}^n\\rightarrow \\{0,1\\}

           \\mathbf{x}\\rightarrow \\mathbf{a}\\cdot\\mathbf{x}+b\\pmod{2}

           (\\mathbf{a}\\in\\{0,1\\}^n, b\\in\\{0,1\\})

        where :math:`(\\cdot)` is the bitwise dot product, that represents the transition-matrix
        elements of the corresponding qubit and ancilla subsystems.

        :param Dict[String, String] bitstring_map: truth-table of the input bitstring map in
        dictionary format
        :return: a dense matrix containing the permutation of the bit strings and a dictionary
        containing the indices of the non-zero elements of the computed permutation matrix as
        key-value-pairs
        :rtype: Tuple[2darray, Dict[String, String]]
        """
        n_bits = len(list(bitstring_map.keys())[0])
        n_ancillas = 1

        # We instantiate an empty matrix of size n_bits + 1 to encode the mapping from n qubits
        # to one ancillas, which explains the additional +1 overhead.
        # To construct the matrix we go through all possible state transitions and pad the index
        # according to all possible states the ancilla-subsystem could be in
        ufunc = np.zeros(shape=(2 ** (n_bits + 1), 2 ** (n_bits + 1)))
        index_mapping_dct = defaultdict(dict)
        for b in range(2**n_ancillas):
            # padding according to ancilla state
            pad_str = np.binary_repr(b, width=1)
            for k, v in bitstring_map.items():
                # add mapping from initial state to the state in the ancilla system.
                # pad_str corresponds to the initial state of the ancilla system.
                index_mapping_dct[pad_str + k] = utils.bitwise_xor(pad_str, v) + k
                # calculate matrix indices that correspond to the transition-matrix-element
                # of the oracle unitary
                i, j = int(pad_str+k, 2), int(utils.bitwise_xor(pad_str, v) + k, 2)
                ufunc[i, j] = 1
        return ufunc, index_mapping_dct

    def _create_bv_circuit(self, bit_map):
        """
        Implementation of the Bernstein-Vazirani Algorithm.

        Given a list of input qubits and an ancilla bit, all initially in the
        :math:`\\vert 0\\rangle` state, create a program that can find :math:`\\vec{a}` with one
        query to the given oracle.

        :param Dict[String, String] bit_map: truth-table of a function for Bernstein-Vazirani with
        the keys being all possible bit vectors strings and the values being the function values
        :rtype: Program
        """
        unitary, _ = self._compute_unitary_oracle_matrix(bit_map)
        full_bv_circuit = pq.Program()

        full_bv_circuit.defgate("BV-ORACLE", unitary)

        # Put ancilla bit into minus state
        full_bv_circuit.inst(X(self.ancilla), H(self.ancilla))

        full_bv_circuit.inst([H(i) for i in self.computational_qubits])
        full_bv_circuit.inst(
            tuple(["BV-ORACLE"] + sorted(self.computational_qubits + [self.ancilla], reverse=True)))
        full_bv_circuit.inst([H(i) for i in self.computational_qubits])
        return full_bv_circuit

    def run(self, cxn, bitstring_map):
        """
        Runs the Bernstein-Vazirani algorithm.

        Given a connection to a QVM or QPU, find the :math:`\\mathbf{a}` and :math:`b` corresponding
        to the function represented by the oracle function that will be constructed from the
        bitstring map.

        :param Connection cxn: connection to the QPU or QVM
        :param Dict[String, String] bitstring_map: a truth table describing the boolean function,
            whose dot-product vector and bias is to be found
        :rtype: BernsteinVazirani
        """

        # initialize all attributes
        self.input_bitmap = bitstring_map
        self.n_qubits = len(list(bitstring_map.keys())[0])
        self.computational_qubits = list(range(self.n_qubits))
        self.ancilla = self.n_qubits  # is the highest index now.

        # construct BV circuit
        self.bv_circuit = self._create_bv_circuit(bitstring_map)

        # find vector by running the full bv circuit
        results = cxn.run_and_measure(self.bv_circuit, self.computational_qubits)
        bv_vector = results[0][::-1]

        # To get the bias term we skip the Walsh-Hadamard transform
        results = cxn.run_and_measure(self.bv_circuit, [self.ancilla])
        bv_bias = results[0][0]

        self.solution = ''.join([str(b) for b in bv_vector]), str(bv_bias)
        return self

    def get_solution(self):
        """
        Returns the solution of the BV algorithm

        :return: a tuple of string corresponding to the dot-product partner vector and the bias term
        :rtype: Tuple[String, String]
        """
        if self.solution is None:
            raise AssertionError("You need to `run` this algorithm first")
        return self.solution

    def check_solution(self):
        """
        Checks if the the found solution correctly reproduces the input.

        :return: True if solution correctly reproduces input bitstring map
        :rtype: Bool
        """
        if self.solution is None:
            raise AssertionError("You need to `run` this algorithm first")
        assert_map = create_bv_bitmap(*self.solution)
        return all([assert_map[k] == v for k, v in self.input_bitmap.items()])



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import pytest
from grove.pyqaoa.maxcut_qaoa import maxcut_qaoa
from grove.pyqaoa.qaoa import QAOA
from grove.pyqaoa.utils import compare_progs
from pyquil.quil import Program
from pyquil.paulis import PauliTerm, PauliSum
from pyquil.gates import H, X, PHASE, CNOT, RZ
import numpy as np
from mock import Mock, patch
import pyquil.api as qvm_mod


def test_pass_hamiltonians():
    ref_ham = [PauliSum([PauliTerm("X", 0, -1.0)]), PauliSum([PauliTerm("X", 1,
                                                              -1.0)])]
    cost_ham = [PauliTerm("I", 0, 0.5) + PauliTerm("Z", 0, -0.5) *
                PauliTerm("Z", 1, 1.0)]
    fakeQVM = Mock()
    inst = QAOA(fakeQVM, 2, steps=1,
                cost_ham=cost_ham, ref_hamiltonian=ref_ham)

    c = inst.cost_ham
    r = inst.ref_ham
    assert isinstance(c, list)
    assert isinstance(r, list)
    assert isinstance(c[0], PauliSum)
    assert isinstance(r[0], PauliSum)
    assert len(c) == 1
    assert len(r) == 2

    with pytest.raises(TypeError):
        QAOA(fakeQVM, 2, steps=1,
             cost_ham=PauliTerm("X", 0, 1.0), ref_hamiltonian=ref_ham,
             rand_seed=42)


def test_hamiltonians():
    test_graph = [(0, 1)]
    p = 1
    inst = maxcut_qaoa(test_graph, steps=p)

    cost_ops, ref_func = inst.cost_ham, inst.ref_ham
    for op in cost_ops:
        for term in op.terms:
            assert(np.isclose(np.abs(term.coefficient), 0.5))

    assert len(ref_func) == 2
    assert len(cost_ops) == 1


def test_param_prog_p1_barbell():
    test_graph = [(0, 1)]
    p = 1
    with patch('grove.pyqaoa.maxcut_qaoa.api', spec=qvm_mod):
        inst = maxcut_qaoa(test_graph, steps=p)

        param_prog = inst.get_parameterized_program()
        trial_prog = param_prog([1.2, 3.4])
        result_prog = Program().inst([H(0), H(1), X(0), PHASE(1.7)(0), X(0),
                                     PHASE(1.7)(0), CNOT(0, 1), RZ(3.4)(1),
                                     CNOT(0, 1), H(0), RZ(-2.4)(0), H(0), H(1),
                                     RZ(-2.4)(1), H(1)])
        compare_progs(trial_prog, result_prog)


def test_psiref_bar_p2():
    bar = [(0, 1)]
    p = 2
    with patch('grove.pyqaoa.maxcut_qaoa.api', spec=qvm_mod):
        inst = maxcut_qaoa(bar, steps=p)

    param_prog = inst.get_parameterized_program()

    # returns are the rotations correct?
    prog = param_prog([1.2, 3.4, 2.1, 4.5])
    result_prog = Program().inst([H(0), H(1),
                                  X(0), PHASE(1.05)(0), X(0), PHASE(1.05)(0),
                                  CNOT(0, 1), RZ(2.1)(1), CNOT(0, 1),
                                  H(0), RZ(-2.4)(0), H(0),
                                  H(1), RZ(-2.4)(1), H(1),
                                  X(0), PHASE(2.25)(0), X(0), PHASE(2.25)(0),
                                  CNOT(0, 1), RZ(4.5)(1), CNOT(0, 1),
                                  H(0), RZ(-6.8)(0), H(0),
                                  H(1), RZ(-6.8)(1), H(1),
                                  ])
    compare_progs(prog, result_prog)


if __name__ == "__main__":
    test_psiref_bar_p2()



import numpy as np

from grove.simon import utils as u


def test_unitary_two_by_two():
    hadamard = np.array([[1., 1.], [1., -1.]])
    hadamard *= 1 / np.sqrt(2)
    assert u.is_unitary(hadamard)


def test_unitary_eight_by_eight():
    matrix = np.zeros(shape=(8, 8))
    one_locations = [(0, 5), (1, 7), (2, 0), (3, 4),
                     (4, 1), (5, 2), (6, 6), (7, 3)]
    for loc in one_locations:
        matrix[loc[0], loc[1]] = 1
    assert u.is_unitary(matrix)


def test_not_unitary_rectangular():
    matrix = np.array([[0, 1, 0], [1, 0, 1]])
    assert not u.is_unitary(matrix)


def test_not_unitary_four_by_four():
    matrix = np.zeros(shape=(4, 4))
    matrix[0, 1] = 1
    matrix[1, 0] = 1
    matrix[2, 2] = 1
    matrix[3, 2] = 1
    assert not u.is_unitary(matrix)


def test_single_one():
    assert u.most_significant_bit(np.array([1])) == 0


def test_single_one_leading_zeroes():
    assert u.most_significant_bit(np.array([0, 1, 0, 0])) == 1


def test_multiple_ones_leading_zeroes():
    assert u.most_significant_bit(np.array([0, 0, 1, 1, 0, 1])) == 2


def test_bit_masking():
    bit_string = '101'
    mask_string = '110'
    assert u.bitwise_xor(bit_string, mask_string) == '011'


def test_binary_back_step():
    A = np.asarray([[1, 1, 0, 1],
                    [0, 1, 1, 0],
                    [0, 0, 1, 1],
                    [0, 0, 0, 1]])
    s = np.asarray([0, 0, 0, 1])

    np.testing.assert_equal(
        u.binary_back_substitute(A, s),
        [1, 1, 0, 1]
    )

    A = np.asarray([[1, 1, 0, 0],
                    [0, 1, 1, 0],
                    [0, 0, 1, 1],
                    [0, 0, 0, 1]])
    s = np.asarray([0, 0, 0, 1])

    np.testing.assert_equal(
        u.binary_back_substitute(A, s),
        [1, 1, 0, 0]
    )

    A = np.asarray([[1, 0, 1, 0],
                    [0, 1, 0, 0],
                    [0, 0, 1, 1],
                    [0, 0, 0, 1]])
    s = np.asarray([0, 0, 0, 1])

    np.testing.assert_equal(
        u.binary_back_substitute(A, s),
        [1, 1, 0, 0]
    )

    A = np.asarray([[1, 0, 1],
                    [0, 1, 0],
                    [0, 0, 1]])
    s = np.asarray([0, 1, 0])

    np.testing.assert_equal(
        u.binary_back_substitute(A, s),
        [0, 1, 0]
    )

    A = np.asarray([[1, 0, 0],
                    [0, 1, 1],
                    [0, 0, 1]])
    s = np.asarray([0, 0, 1])

    np.testing.assert_equal(
        u.binary_back_substitute(A, s),
        [1, 1, 0]
    )



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Class to represent a Fenwick tree.
"""


class FenwickNode:
    """Fenwick Tree node."""
    parent = None
    children = None
    index = None

    def __init__(self, parent, children, index=None):
        """Fenwick Tree node. Single parent and multiple children.
        
        :param FenwickNode parent: a parent node
        :param list(FenwickNode) children: a list of children nodes
        :param int index: node label
        """
        self.children = children
        self.parent = parent
        self.index = index

    def get_ancestors(self):
        """Returns a list of ancestors of the node. Ordered from the earliest.

        :return: node's ancestors, ordered from most recent
        :rtype: list(FenwickNode)
        """
        node = self
        ancestor_list = []
        while node.parent is not None:
            ancestor_list.append(node.parent)
            node = node.parent

        return ancestor_list


class FenwickTree:
    """Recursive implementation of the Fenwick tree.
    Please see Subsection B.2. of Operator Locality in Quantum
    Simulation of Fermionic Models (arXiv:1701.07072) for
    a reference to the update set (U), the parity set (P) and the
    children set (F) sets of the Fenwick.
    """
    # Root node.
    root = None

    def __init__(self, n_qubits):
        """
        Builds a Fenwick tree on n_qubits qubits.

        :param int n_qubits: number of qubits in the system
        """
        self.nodes = [FenwickNode(None, []) for _ in range(n_qubits)]

        if n_qubits > 0:
            self.root = self.nodes[n_qubits - 1]
            self.root.index = n_qubits - 1

        def fenwick(left, right, parent):
            """
            This inner function is used to build the Fenwick tree on nodes
            recursively. See Algorithm 1 in the paper.

            :param int left: left boundary of range
            :param int right: right boundary of range
            :param FenwickNode parent: parent node
            """
            if left >= right:
                return
            else:
                pivot = (left + right) >> 1
                child = self.nodes[pivot]

                # The circle of life:
                # Parent has child.
                # Child becomes parent.
                child.index = pivot
                parent.children.append(child)
                child.parent = parent

                # Recurse to left and to right.
                fenwick(left, pivot, child)
                fenwick(pivot + 1, right, parent)

        # Builds structure on nodes.
        fenwick(0, n_qubits - 1, self.root)

    def get_node(self, j):
        """Returns the node at j in the qubit register. Wrapper.

        :param int j: fermionic site index

        :return: the node at j
        :rtype: FenwickNode
        """
        return self.nodes[j]

    def get_update_set(self, j):
        """The set of all ancestors of j, (the update set U from the paper).

        :param int j: fermionic site index

        :return: ancestors of j, ordered from most recent
        :rtype: list(FenwickNode)
        """
        node = self.get_node(j)
        return node.get_ancestors()

    def get_children_set(self, j):
        """Returns the set of children of j-th site.

        :param int j: fermionic site index

        :return: children of j, ordered from lowest index
        :rtype: list(FenwickNode)
        """
        node = self.get_node(j)
        return node.children

    def get_remainder_set(self, j):
        """Return the set of children with indices less than j of all ancestors
        of j. The set C from (arXiv:1701.07072).

        :param int j: fermionic site index

        :return: children of j-ancestors, with indices less than j
        :rtype: list(FenwickNode)
        """
        result = []
        ancestors = self.get_update_set(j)

        # This runs in O(log(N)log(N)) where N is the number of qubits.
        for a in ancestors:
            for c in a.children:
                if c.index < j:
                    result.append(c)

        return result

    def get_parity_set(self, j):
        """Returns the union of the remainder set with children set. Coincides
        with the parity set of Tranter et al.

        :param int j: fermionic site index

        :return: union of remainder and parity set for node with index j
        :rtype: list(FenwickNode)
        """
        return self.get_remainder_set(j) + self.get_children_set(j)



import numpy as np
from pyquil.quil import Program

from grove.tests.utils.utils_for_testing import non_action_insts, prog_len
from grove.utils.utility_programs import ControlledProgramBuilder

SIGMA_Z = np.array([[1, 0], [0, -1]])
SIGMA_Z_NAME = "Z"


def test_1_qubit_control():
    prog = Program()
    qubit = prog.alloc()
    control_qubit = prog.alloc()
    prog += (ControlledProgramBuilder()
             .with_controls([control_qubit])
             .with_target(qubit)
             .with_operation(SIGMA_Z)
             .with_gate_name(SIGMA_Z_NAME).build())
    # This should be one "CZ" instruction, from control_qubit to qubit.
    assert prog_len(prog) == 1
    prog.synthesize()
    instruction = non_action_insts(prog)[0]
    assert instruction[1].operator_name == (ControlledProgramBuilder()
                                            .format_gate_name("C", SIGMA_Z_NAME))
    assert instruction[1].arguments == [control_qubit, qubit]


def double_control_test(instructions, target_qubit, control_qubit_one, control_qubit_two):
    """A list of asserts testing the simple case of a double controlled Z gate. Used in the next
     two tests."""
    cpg = ControlledProgramBuilder()
    sqrt_z = cpg.format_gate_name("SQRT", SIGMA_Z_NAME)
    assert instructions[0][1].operator_name == (cpg.format_gate_name("C", sqrt_z))
    assert instructions[0][1].arguments == [control_qubit_two, target_qubit]

    assert instructions[1][1].operator_name == cpg.format_gate_name("C", "NOT")
    assert instructions[1][1].arguments == [control_qubit_one, control_qubit_two]

    assert instructions[2][1].operator_name == cpg.format_gate_name("C", sqrt_z) + '-INV'
    assert instructions[2][1].arguments == [control_qubit_two, target_qubit]

    assert instructions[3][1].operator_name == cpg.format_gate_name("C", "NOT")
    assert instructions[3][1].arguments == [control_qubit_one, control_qubit_two]

    assert instructions[4][1].operator_name == cpg.format_gate_name("C", sqrt_z)
    assert instructions[4][1].arguments == [control_qubit_one, target_qubit]


def test_recursive_builder():
    """Here we test the _recursive_builder in ControlledProgramBuilder individually."""
    control_qubit_one = 1
    control_qubit_two = 2
    target_qubit = 3
    cpg = (ControlledProgramBuilder()
           .with_controls([control_qubit_one, control_qubit_two])
           .with_target(target_qubit)
           .with_operation(SIGMA_Z)
           .with_gate_name(SIGMA_Z_NAME))
    prog = cpg._recursive_builder(cpg.operation,
                                  cpg.gate_name,
                                  cpg.control_qubits,
                                  cpg.target_qubit)
    instructions = non_action_insts(prog)
    # Run tests
    double_control_test(instructions, target_qubit, control_qubit_one, control_qubit_two)


def test_2_qubit_control():
    """Test that ControlledProgramBuilder builds the program correctly all the way through."""
    prog = Program()
    qubit = prog.alloc()
    control_qubit_one = prog.alloc()
    control_qubit_two = prog.alloc()
    prog += (ControlledProgramBuilder()
             .with_controls([control_qubit_one, control_qubit_two])
             .with_target(qubit)
             .with_operation(SIGMA_Z)
             .with_gate_name(SIGMA_Z_NAME).build())
    # This should be one "CZ" instruction, from control_qubit to qubit.
    assert prog_len(prog) == 5
    prog.synthesize()
    instructions = non_action_insts(prog)
    # Run tests
    double_control_test(instructions, qubit, control_qubit_one, control_qubit_two)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from pyquil.gates import *
from grove.qft.fourier import inverse_qft
import pyquil.quil as pq
from grove.pyqaoa.utils import compare_progs # This would be nice to have in
                                             # general purpose Util module


def test_simple_inverse_qft():
    
    trial_prog = pq.Program()
    trial_prog.inst(X(0))
    trial_prog = trial_prog + inverse_qft([0])
    
    result_prog = pq.Program().inst([X(0), H(0)])
    
    compare_progs(trial_prog, result_prog)


def test_multi_qubit_qft():
    
    trial_prog = pq.Program()
    trial_prog.inst(X(0), X(1), X(2))
    trial_prog = trial_prog + inverse_qft([0, 1, 2])
    
    result_prog = pq.Program().inst([X(0), X(1), X(2),
                                     SWAP(0, 2), H(0),
                                     CPHASE(-1.5707963267948966, 0, 1),
                                     CPHASE(-0.7853981633974483, 0, 2),
                                     H(1), CPHASE(-1.5707963267948966, 1, 2),
                                     H(2)])
    
    compare_progs(trial_prog, result_prog)



from itertools import product

from grove.alpha.fermion_transforms.jwtransform import JWTransform


def test_create():
    """
    Testing creation operator produces 0.5 * (X - iY)
    """
    jw = JWTransform()
    creation = jw.create(5)
    # we really need PauliSum and PauliTerm logical compare
    assert '(0.5+0j)*Z0*Z1*Z2*Z3*Z4*X5 + -0.5j*Z0*Z1*Z2*Z3*Z4*Y5' == creation.__str__()


def test_annihilation():
    """
    Testing creation operator produces 0.5 * (X - iY)
    """
    jw = JWTransform()
    annihilation = jw.kill(5)
    # we really need PauliSum and PauliTerm logical compare
    assert '(0.5+0j)*Z0*Z1*Z2*Z3*Z4*X5 + 0.5j*Z0*Z1*Z2*Z3*Z4*Y5' == annihilation.__str__()


def test_hopping():
    """
    Hopping term tests
    """
    jw = JWTransform()
    hopping = jw.create(2)*jw.kill(0) + jw.create(0)*jw.kill(2)
    assert '(0.5+0j)*X0*Z1*X2 + (0.5+0j)*Y0*Z1*Y2' == hopping.__str__()


def test_multi_ops():
    """
    test construction of Paulis for product of second quantized operators
    """
    jw = JWTransform()
    # test on one particle density matrix
    for p, q, in product(range(6), repeat=2):
        truth = jw.create(p)*jw.kill(q)
        prod_ops_out = jw.product_ops([p, q], [-1, 1])
        assert truth.__str__() == prod_ops_out.__str__()

    # test on two particle density matrix
    for p, q, r, s in product(range(4), repeat=4):
        truth = jw.create(p)*jw.create(q)*jw.kill(s)*jw.kill(r)
        prod_ops_out = jw.product_ops([p, q, s, r], [-1, -1, 1, 1])
        assert truth.__str__() == prod_ops_out.__str__()



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import pyquil.quil as pq
from pyquil.gates import H
import numpy as np
from math import log
from grove.qft.fourier import qft


def controlled(m):
    """
    Make a one-qubit-controlled version of a matrix.

    :param m: (numpy.ndarray) A matrix.
    :return: A controlled version of that matrix.
    """
    rows, cols = m.shape
    assert rows == cols
    n = rows
    I = np.eye(n)
    Z = np.zeros((n, n))
    controlled_m = np.bmat([[I, Z],
                            [Z, m]])
    return controlled_m


def phase_estimation(U, accuracy, reg_offset=0):
    """
    Generate a circuit for quantum phase estimation.

    :param U: (numpy.ndarray) A unitary matrix.
    :param accuracy: (int) Number of bits of accuracy desired.
    :param reg_offset: (int) Where to start writing measurements (default 0).
    :return: A Quil program to perform phase estimation.
    """
    assert isinstance(accuracy, int)
    rows, cols = U.shape
    m = int(log(rows, 2))
    output_qubits = range(0, accuracy)
    U_qubits = range(accuracy, accuracy + m)
    p = pq.Program()

    # Hadamard initialization
    for i in output_qubits:
        p.inst(H(i))
    # Controlled unitaries
    for i in output_qubits:
        if i > 0:
            U = np.dot(U, U)
        cU = controlled(U)
        name = "CONTROLLED-U{0}".format(2 ** i)
        # define the gate
        p.defgate(name, cU)
        # apply it
        p.inst((name, i) + tuple(U_qubits))
    # Compute the QFT
    p = p + qft(output_qubits)
    # Perform the measurements
    for i in output_qubits:
        p.measure(i, reg_offset + i)

    return p


if __name__ == '__main__':
    import pyquil.api as api
    qvm = api.SyncConnection()
    X = np.asarray([[0.0, 1.0], [1.0, 0.0]])
    Y = np.asarray([[0.0, -1.0j], [1.0j, 0.0]])
    Rx = np.exp(X * np.pi / 8)
    Ry = np.exp(Y * np.pi / 16)
    U = np.kron(Rx, Ry)
    p = phase_estimation(U, 3)
    print(p)
    print(qvm.run(p, range(3+2), 10))
    print(qvm.wavefunction(p))



"""Module containing utilities that will eventually be implemented in pyquil."""

from pyquil.quilbase import (ACTION_INSTANTIATE_QUBIT,
                             ACTION_RELEASE_QUBIT)
from pyquil.quilbase import Gate, Qubit
from pyquil.resource_manager import merge_resource_managers, ResourceManager
# TODO Implement this in pyQuil.


def prog_equality(prog1, prog2):
    """Tests to see if two programs are equal based on their out methods

    :param prog1: The first Program to compare
    :param prog2: The second program to compare
    """
    return prog1.out() == prog2.out()


def non_action_insts(prog):
    """Returns quantum gates, and classical control flow of the program, ignoring qubit allocation,
     and qubit release.

    :param prog: The program to return instructions for.
    :return: The instructions in the program, ignoring qubit allocation and deallocation.
    :rtype: list
    """
    insts = []
    for action in prog:
        if action[0] not in (ACTION_INSTANTIATE_QUBIT, ACTION_RELEASE_QUBIT):
            insts.append(action)
    return insts


def prog_len(prog):
    """Returns the length of of a program, defined as the number of instructions.

    :return: The length of prog.
    :rtype: int
    """
    length = 0
    for action in prog:
        if action[0] == ACTION_INSTANTIATE_QUBIT:
            continue
        length += 1
    return length


def synthesize_programs(*progs):
    """Synthesizes programs together, paying attention to shared qubits. This breaks if synthesize
     is called after, very fragile.

    :param progs: List of Programs.
    :return: None
    :rtype: NoneType"""
    def get_allocated_qubits(*progs):
        qubits_to_progs = {}
        for prog in progs:
            for inst in prog:
                if inst[0] == ACTION_INSTANTIATE_QUBIT:
                    qubits_to_progs[inst[1]] = [prog]
        return qubits_to_progs
    qubits_to_progs = get_allocated_qubits(*progs)
    for prog in progs:
        for inst in prog:
            if isinstance(inst[1], Gate):
                for arg in inst[1].arguments:
                    if isinstance(arg, Qubit):
                        qubits_to_progs[arg].append(prog)

    equiv_classes = list(qubits_to_progs.values())
    repeat = True
    while repeat:
        repeat = False
        for i, equiv_class in enumerate(equiv_classes):
            for j, other_equiv_class in enumerate(equiv_classes):
                if j <= i:
                    continue
                if set(equiv_class).intersection(set(other_equiv_class)):
                    equiv_classes[i] = set(equiv_class).union(other_equiv_class)
                    equiv_classes.remove(other_equiv_class)
                    repeat = True

    for equiv_class in equiv_classes:
        rm = ResourceManager()
        for prog in equiv_class:
            rm = merge_resource_managers(rm, prog.resource_manager)
        for prog in equiv_class:
            prog.resource_manager = rm
            prog.actions = [inst for inst in prog if inst[0] != ACTION_INSTANTIATE_QUBIT]
        for qubit in rm.live_qubits:
            for prog in progs:
                prog.actions = [(ACTION_INSTANTIATE_QUBIT, qubit)] + prog.actions



"""Tests for utils"""

import pytest

from grove.alpha.arbitrary_state.arbitrary_state import *
from grove.pyqaoa.utils import compare_progs


@pytest.mark.skip(reason="Must add support for Forest connections in testing")
class TestCreateArbitraryState(object):
    def test_state_generation_simple_length_two(self):
        _state_generation_test_helper([1 + 1j, 2])

    def test_state_generation_simple_length_four(self):
        _state_generation_test_helper([1, 2, 3, 4])

    def test_state_generation_complex_length_five(self):
        _state_generation_test_helper([1 + 2j, 3 + 4j, -1 - 5j, 6 - 8j, 7])

    def test_state_generation_complex_length_eight(self):
        _state_generation_test_helper(
            [1 + 2j, 3 + 4j, -1 - 5j, 6 - 8j, 7, 2, 0.5j, 2])

    def test_state_generation_complex_length_huge(self):
        _state_generation_test_helper(np.array(range(50)) * 1j)

    def test_padded_zeros(self):
        _state_generation_test_helper([1, 0, 0])

    def test_single_one(self):
        _state_generation_test_helper([1], [4])

    def test_long_padded_zeros(self):
        _state_generation_test_helper([0.5j, 0.5, 0, 1, 0, 0, 0, 0, 0],
                                      range(3, 7))

    def test_forward_padded_zeros(self):
        _state_generation_test_helper([0, 0, 1])


class TestUniformlyControlledRotationMatrix(object):
    def test_one_control(self):
        expected = 1 / 2. * np.array([[1, 1],
                                      [1, -1]])
        actual = get_uniformly_controlled_rotation_matrix(1)
        assert np.allclose(expected, actual)

    def test_two_controls(self):
        expected = 1 / 4. * np.array([[1, 1, 1, 1],
                                      [1, -1, 1, -1],
                                      [1, -1, -1, 1],
                                      [1, 1, -1, -1]])
        actual = get_uniformly_controlled_rotation_matrix(2)

        assert np.allclose(expected, actual)

    def test_three_controls(self):
        expected = 1 / 8. * np.array([[1, 1, 1, 1, 1, 1, 1, 1],
                                      [1, -1, 1, -1, 1, -1, 1, -1],
                                      [1, -1, -1, 1, 1, -1, -1, 1],
                                      [1, 1, -1, -1, 1, 1, -1, -1],
                                      [1, 1, -1, -1, -1, -1, 1, 1],
                                      [1, -1, -1, 1, -1, 1, 1, -1],
                                      [1, -1, 1, -1, -1, 1, -1, 1],
                                      [1, 1, 1, 1, -1, -1, -1, -1]])
        actual = get_uniformly_controlled_rotation_matrix(3)

        assert np.allclose(expected, actual)


class TestUniformlyControlledCNOTPositions(object):
    def test_one_control(self):
        expected = [1, 1]
        actual = get_cnot_control_positions(1)
        assert expected == actual

    def test_two_controls(self):
        expected = [1, 2, 1, 2]
        actual = get_cnot_control_positions(2)
        assert expected == actual

    def test_three_controls(self):
        expected = [1, 2, 1, 3, 1, 2, 1, 3]
        actual = get_cnot_control_positions(3)
        assert expected == actual

    def test_four_controls(self):
        expected = [1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1, 4]
        actual = get_cnot_control_positions(4)
        assert expected == actual


class TestGetRotationParameters(object):
    def test_length_four_coefficients(self):
        phases = [np.pi / 2, np.pi / 6, -np.pi / 7, np.pi / 8]
        magnitudes = [0.1, 0.2, 0.3, 0.4]
        z_thetas, y_thetas, new_phases, new_magnitudes = \
            get_rotation_parameters(phases, magnitudes)
        expected_z_thetas = [np.pi / 3, -np.pi * 15 / 56]
        expected_y_thetas = [2 * np.arcsin(-0.1 / np.sqrt(0.1)),
                             2 * np.arcsin(-0.1 / np.sqrt(0.5))]
        expected_new_phases = [np.pi / 3, -np.pi / 112]
        expected_new_magnitudes = [np.sqrt(0.025), np.sqrt(0.125)]

        assert np.allclose(expected_z_thetas, z_thetas)
        assert np.allclose(expected_y_thetas, y_thetas)
        assert np.allclose(expected_new_phases, new_phases)
        assert np.allclose(expected_new_magnitudes, new_magnitudes)

    def test_length_eight_coefficients(self):
        phases = [-np.pi / 6, -np.pi / 6, np.pi / 13, -np.pi / 7,
                  -np.pi / 9, np.pi / 2, np.pi / 4, np.pi / 5]
        magnitudes = [0.7, 0.3, 0.25, 0.6,
                      0.84, 0.84, 0.05, 0.4]
        z_thetas, y_thetas, new_phases, new_magnitudes = \
            get_rotation_parameters(phases, magnitudes)
        expected_z_thetas = [0, np.pi * 20 / 91, -np.pi * 11 / 18, np.pi / 20]
        expected_y_thetas = [2 * np.arcsin(0.4 / np.sqrt(1.16)),
                             2 * np.arcsin(-0.35 / np.sqrt(0.845)),
                             0,
                             2 * np.arcsin(-0.35 / np.sqrt(0.325))]
        expected_new_phases = [-np.pi / 6, -np.pi * 3 / 91, np.pi * 7 / 36,
                               np.pi * 9 / 40]
        expected_new_magnitudes = [np.sqrt(0.29), np.sqrt(0.21125),
                                   0.84, np.sqrt(0.08125)]

        assert np.allclose(expected_z_thetas, z_thetas)
        assert np.allclose(expected_y_thetas, y_thetas)
        assert np.allclose(expected_new_phases, new_phases)
        assert np.allclose(expected_new_magnitudes, new_magnitudes)


class TestGetReversedUnificationProgram(object):
    def test_length_four_phase_rotations(self):
        angles = [np.pi / 4, np.pi / 11, np.pi / 7, -np.pi / 18]
        control_indices = get_cnot_control_positions(2)
        controls = [3, 8]
        target = 0
        reverse_prog = \
            get_reversed_unification_program(angles, control_indices,
                                             target, controls, "phase")

        expected_prog = pq.Program().inst(CNOT(8, 0)) \
            .inst(RZ(np.pi / 18, 0)) \
            .inst(CNOT(3, 0)) \
            .inst(RZ(-np.pi / 7, 0)) \
            .inst(CNOT(8, 0)) \
            .inst(RZ(-np.pi / 11, 0)) \
            .inst(CNOT(3, 0)) \
            .inst(RZ(-np.pi / 4, 0))

        compare_progs(reverse_prog, expected_prog)

    def test_length_eight_magnitude_rotations(self):
        angles = [0, - np.pi / 12, np.pi / 15, -np.pi / 6,
                  np.pi / 13, np.pi / 10, -np.pi / 3, np.pi / 4]
        control_indices = get_cnot_control_positions(3)
        controls = [1, 6, 2]
        target = 4
        reverse_prog = \
            get_reversed_unification_program(angles, control_indices,
                                             target, controls, "magnitude")

        expected_prog = pq.Program().inst(CNOT(2, 4)) \
            .inst(RY(-np.pi / 4, 4)) \
            .inst(CNOT(1, 4)) \
            .inst(RY(np.pi / 3, 4)) \
            .inst(CNOT(6, 4)) \
            .inst(RY(-np.pi / 10, 4)) \
            .inst(CNOT(1, 4)) \
            .inst(RY(-np.pi / 13, 4)) \
            .inst(CNOT(2, 4)) \
            .inst(RY(np.pi / 6, 4)) \
            .inst(CNOT(1, 4)) \
            .inst(RY(- np.pi / 15, 4)) \
            .inst(CNOT(6, 4)) \
            .inst(RY(np.pi / 12, 4)) \
            .inst(CNOT(1, 4))

        compare_progs(reverse_prog, expected_prog)


def _state_generation_test_helper(v, qubits=None):
    # encode vector in quantum state
    p = create_arbitrary_state(v, qubits)
    qvm = SyncConnection()
    wf, _ = qvm.wavefunction(p)

    # normalize and pad with zeros
    v_norm = v / np.linalg.norm(v)
    while len(v_norm) < len(wf.amplitudes):
        v_norm = np.append(v_norm, 0)

    # wavefunction amplitudes should resemble vector
    for pair in zip(v_norm, wf.amplitudes):
        assert np.allclose(*pair)
